from unittest.mock import ANY

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    JUST_PANDAS_DATA_SOURCES,
    NON_SQL_DATA_SOURCES,
    SQL_DATA_SOURCES,
)

DIFFERENT_COL = "some_are_different"
SAME_COL = "all_the_same"

DATA = pd.DataFrame(
    {
        SAME_COL: ["FOO", "BAR", "BAZ", None],
        DIFFERENT_COL: ["FOOD", "BAR", "BAZ", None],
    }
)


@parameterize_batch_for_data_sources(data_source_configs=SQL_DATA_SOURCES, data=DATA)
def test_success_complete__sql(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValueLengthsToEqual(column=SAME_COL, value=3)
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {
        "element_count": 4,
        "unexpected_count": 0,
        "unexpected_percent": 0.0,
        "partial_unexpected_list": [],
        "missing_count": 1,
        "missing_percent": 25.0,
        "unexpected_percent_total": 0.0,
        "unexpected_percent_nonmissing": 0.0,
        "partial_unexpected_counts": [],
        "unexpected_list": [],
        "unexpected_index_query": ANY,
    }


@parameterize_batch_for_data_sources(data_source_configs=NON_SQL_DATA_SOURCES, data=DATA)
def test_success_complete__non_sql(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValueLengthsToEqual(column=SAME_COL, value=3)
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValueLengthsToEqual(column=SAME_COL, value=3),
            id="exact_match",
        ),
        pytest.param(
            gxe.ExpectColumnValueLengthsToEqual(column=SAME_COL, value=3, mostly=0.75),
            id="with_mostly",
        ),
        pytest.param(
            gxe.ExpectColumnValueLengthsToEqual(column=DIFFERENT_COL, value=3, mostly=0.1),
            id="different_lengths_with_mostly",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnValueLengthsToEqual
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValueLengthsToEqual(column=DIFFERENT_COL, value=3),
            id="wrong_length",
        ),
        pytest.param(
            gxe.ExpectColumnValueLengthsToEqual(column=DIFFERENT_COL, value=3, mostly=0.9),
            id="mostly_too_high",
        ),
        pytest.param(
            gxe.ExpectColumnValueLengthsToEqual(column=SAME_COL, value=4, mostly=0.1),
            id="no_matches",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnValueLengthsToEqual
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success



from .store import Store  # isort:skip
from .store_backend import (  # isort:skip
    StoreBackend,
)
from .gx_cloud_store_backend import GXCloudStoreBackend  # isort:skip
from .tuple_store_backend import (  # isort:skip
    TupleFilesystemStoreBackend,
    TupleGCSStoreBackend,
    TupleS3StoreBackend,
    TupleStoreBackend,
    TupleAzureBlobStoreBackend,
)
from .database_store_backend import DatabaseStoreBackend  # isort:skip
from .inline_store_backend import InlineStoreBackend  # isort:skip
from .in_memory_store_backend import InMemoryStoreBackend  # isort:skip
from .configuration_store import ConfigurationStore  # isort:skip
from .checkpoint_store import CheckpointStore  # isort:skip
from .metric_store import (  # isort:skip
    MetricStore,
)
from .expectations_store import ExpectationsStore  # isort:skip
from .validation_results_store import ValidationResultsStore  # isort:skip
from .query_store import SqlAlchemyQueryStore  # isort:skip
from .html_site_store import HtmlSiteStore  # isort:skip
from .datasource_store import DatasourceStore  # isort:skip
from .data_context_store import DataContextStore  # isort:skip
from .data_asset_store import DataAssetStore  # isort:skip
from .validation_definition_store import ValidationDefinitionStore  # isort:skip



from .base import DataSourceTestConfig
from .big_query import BigQueryDatasourceTestConfig
from .databricks import DatabricksDatasourceTestConfig
from .mssql import MSSQLDatasourceTestConfig
from .mysql import MySQLDatasourceTestConfig
from .pandas_data_frame import PandasDataFrameDatasourceTestConfig
from .pandas_filesystem_csv import PandasFilesystemCsvDatasourceTestConfig
from .postgres import PostgreSQLDatasourceTestConfig
from .redshift import RedshiftDatasourceTestConfig
from .snowflake import SnowflakeDatasourceTestConfig
from .spark_filesystem_csv import SparkFilesystemCsvDatasourceTestConfig
from .sqlite import SqliteDatasourceTestConfig



from __future__ import annotations

import logging
import traceback
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    TypedDict,
    Union,
)

from tqdm.auto import tqdm

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility.typing_extensions import override
from great_expectations.expectations.registry import get_metric_provider
from great_expectations.validator.exception_info import ExceptionInfo
from great_expectations.validator.metric_configuration import (
    MetricConfiguration,
    MetricConfigurationID,
)

if TYPE_CHECKING:
    from great_expectations.core import IDDict
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.expectations.metrics.metric_provider import MetricProvider
    from great_expectations.validator.computed_metric import MetricValue
    from great_expectations.validator.metrics_calculator import (
        _AbortedMetricsInfoDict,
    )


class MetricsCalculatorErrorResultValue(TypedDict):
    metric_configuration: MetricConfiguration
    exception_info: ExceptionInfo
    num_failures: int


__all__ = [
    "ExpectationValidationGraph",
    "MetricConfiguration",
    "MetricEdge",
    "MetricValue",
    "ValidationGraph",
]

logger = logging.getLogger(__name__)
logging.captureWarnings(True)

MAX_METRIC_COMPUTATION_RETRIES: int = 3


class MetricEdge:
    def __init__(
        self, left: MetricConfiguration, right: Optional[MetricConfiguration] = None
    ) -> None:
        self._left = left
        self._right = right

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    @property
    def id(self):
        if self.right:
            return self.left.id, self.right.id

        return self.left.id, None

    def __repr__(self):  # type: ignore[explicit-override] # FIXME
        return f"<{self._left.__repr__()}|{self._right.__repr__()}>"


class ValidationGraph:
    def __init__(
        self,
        execution_engine: ExecutionEngine,
        edges: Optional[List[MetricEdge]] = None,
    ) -> None:
        self._execution_engine = execution_engine

        if edges:
            self._edges = edges
        else:
            self._edges = []

        self._edge_ids = {edge.id for edge in self._edges}

    @override
    def __eq__(self, other) -> bool:
        """Supports comparing two "ValidationGraph" objects."""
        return self.edge_ids == other.edge_ids

    @override
    def __hash__(self) -> int:
        return hash(frozenset(self.edge_ids))

    @property
    def edges(self) -> List[MetricEdge]:
        """Returns "MetricEdge" objects, contained within this "ValidationGraph" object (as list)."""  # noqa: E501 # FIXME CoP
        return self._edges

    @property
    def edge_ids(self) -> Set[Tuple[str, str]]:
        """Returns "MetricEdge" objects, contained within this "ValidationGraph" object (as set of two-tuples)."""  # noqa: E501 # FIXME CoP
        return {edge.id for edge in self._edges}

    def add(self, edge: MetricEdge) -> None:
        """Adds supplied "MetricEdge" object to this "ValidationGraph" object (if not already present)."""  # noqa: E501 # FIXME CoP
        if edge.id not in self._edge_ids:
            self._edges.append(edge)
            self._edge_ids.add(edge.id)

    def build_metric_dependency_graph(
        self,
        metric_configuration: MetricConfiguration,
        runtime_configuration: Optional[dict] = None,
    ) -> None:
        """
        Obtain domain and value keys for metrics and proceeds to add these metrics to the validation graph
        until all metrics have been added.

        Args:
            metric_configuration: Desired MetricConfiguration object to be resolved.
            runtime_configuration: Additional run-time settings (see "Validator.DEFAULT_RUNTIME_CONFIGURATION").
        """  # noqa: E501 # FIXME CoP

        metric_impl_klass: MetricProvider
        metric_provider: Callable
        (
            metric_impl_klass,
            _metric_provider,
        ) = self.set_metric_configuration_default_kwargs_if_absent(
            metric_configuration=metric_configuration
        )

        metric_dependencies = metric_impl_klass.get_evaluation_dependencies(
            metric=metric_configuration,
            execution_engine=self._execution_engine,
            runtime_configuration=runtime_configuration,
        )

        if len(metric_dependencies) == 0:
            self.add(
                MetricEdge(
                    left=metric_configuration,
                )
            )
        else:
            metric_configuration.metric_dependencies = metric_dependencies
            for metric_dependency in metric_dependencies.values():
                # TODO: <Alex>In the future, provide a more robust cycle detection mechanism.</Alex>
                if metric_dependency.id == metric_configuration.id:
                    logger.warning(
                        f"Metric {metric_configuration.id!s} has created a circular dependency"
                    )
                    continue
                self.add(
                    MetricEdge(
                        left=metric_configuration,
                        right=metric_dependency,
                    )
                )
                self.build_metric_dependency_graph(
                    metric_configuration=metric_dependency,
                    runtime_configuration=runtime_configuration,
                )

    def set_metric_configuration_default_kwargs_if_absent(
        self, metric_configuration: MetricConfiguration
    ) -> Tuple[MetricProvider, Callable]:
        """
        Updates "metric_domain_kwargs" and/or "metric_value_kwargs" of "MetricConfiguration" with defualts (if needed).
        """  # noqa: E501 # FIXME CoP
        metric_impl_klass: MetricProvider
        metric_provider: Callable
        metric_impl_klass, metric_provider = get_metric_provider(
            metric_name=metric_configuration.metric_name,
            execution_engine=self._execution_engine,
        )
        self._set_default_metric_kwargs_if_absent(
            default_kwarg_values=metric_impl_klass.default_kwarg_values,
            metric_kwargs=metric_configuration.metric_domain_kwargs,
            keys=metric_impl_klass.domain_keys,
        )
        self._set_default_metric_kwargs_if_absent(
            default_kwarg_values=metric_impl_klass.default_kwarg_values,
            metric_kwargs=metric_configuration.metric_value_kwargs,
            keys=metric_impl_klass.value_keys,
        )
        return metric_impl_klass, metric_provider

    def resolve(
        self,
        runtime_configuration: Optional[dict] = None,
        min_graph_edges_pbar_enable: int = 0,
        # Set to low number (e.g., 3) to suppress progress bar for small graphs.
        show_progress_bars: bool = True,
    ) -> Tuple[
        Dict[MetricConfigurationID, MetricValue],
        _AbortedMetricsInfoDict,
    ]:
        resolved_metrics: Dict[MetricConfigurationID, MetricValue] = {}

        # updates graph with aborted metrics
        aborted_metrics_info: _AbortedMetricsInfoDict = self._resolve(
            metrics=resolved_metrics,
            runtime_configuration=runtime_configuration,
            min_graph_edges_pbar_enable=min_graph_edges_pbar_enable,
            show_progress_bars=show_progress_bars,
        )

        return resolved_metrics, aborted_metrics_info

    def _resolve(  # noqa: C901, PLR0912, PLR0915 # FIXME CoP
        self,
        metrics: Dict[MetricConfigurationID, MetricValue],
        runtime_configuration: Optional[dict] = None,
        min_graph_edges_pbar_enable: int = 0,  # Set to low number (e.g., 3) to suppress progress bar for small graphs.  # noqa: E501 # FIXME CoP
        show_progress_bars: bool = True,
    ) -> _AbortedMetricsInfoDict:
        if metrics is None:
            metrics = {}

        if runtime_configuration is None:
            runtime_configuration = {}

        if runtime_configuration.get("catch_exceptions", True):
            catch_exceptions = True
        else:
            catch_exceptions = False

        failed_metric_info: _AbortedMetricsInfoDict = {}
        aborted_metrics_info: _AbortedMetricsInfoDict = {}

        ready_metrics: Set[MetricConfiguration]
        needed_metrics: Set[MetricConfiguration]

        exception_info: ExceptionInfo

        progress_bar: Optional[tqdm] = None

        done: bool = False
        while not done:
            ready_metrics, needed_metrics = self._parse(metrics=metrics)

            # Check to see if the user has disabled progress bars
            disable = not show_progress_bars
            if len(self.edges) < min_graph_edges_pbar_enable:
                disable = True

            if progress_bar is None:
                # noinspection PyProtectedMember,SpellCheckingInspection
                progress_bar = tqdm(
                    total=len(ready_metrics) + len(needed_metrics),
                    desc="Calculating Metrics",
                    disable=disable,
                )
            progress_bar.update(0)
            progress_bar.refresh()

            computable_metrics = set()

            for metric in ready_metrics:
                if (
                    metric.id in failed_metric_info
                    and failed_metric_info[metric.id]["num_failures"]
                    >= MAX_METRIC_COMPUTATION_RETRIES
                ):
                    aborted_metrics_info[metric.id] = failed_metric_info[metric.id]
                else:
                    computable_metrics.add(metric)

            try:
                # Access "ExecutionEngine.resolve_metrics()" method, to resolve missing "MetricConfiguration" objects.  # noqa: E501 # FIXME CoP
                metrics.update(
                    self._execution_engine.resolve_metrics(
                        metrics_to_resolve=computable_metrics,
                        metrics=metrics,
                        runtime_configuration=runtime_configuration,
                    )
                )
                progress_bar.update(len(computable_metrics))
                progress_bar.refresh()
            except gx_exceptions.MetricResolutionError as err:
                if catch_exceptions:
                    exception_traceback = traceback.format_exc()
                    exception_message = str(err)
                    exception_info = ExceptionInfo(
                        exception_traceback=exception_traceback,
                        exception_message=exception_message,
                    )
                    for failed_metric in err.failed_metrics:
                        if failed_metric.id in failed_metric_info:
                            failed_metric_info[failed_metric.id]["num_failures"] += 1
                            failed_metric_info[failed_metric.id]["exception_info"] = exception_info
                        else:
                            failed_metric_info[failed_metric.id] = (
                                MetricsCalculatorErrorResultValue(
                                    metric_configuration=failed_metric,
                                    exception_info=exception_info,
                                    num_failures=1,
                                )
                            )

                else:
                    raise err  # noqa: TRY201 # FIXME CoP
            except Exception as e:
                if catch_exceptions:
                    logger.error(  # noqa: TRY400 # FIXME CoP
                        f"""Caught exception {e!s} while trying to resolve a set of {len(ready_metrics)} metrics; aborting graph resolution."""  # noqa: E501 # FIXME CoP
                    )
                    done = True
                else:
                    raise e  # noqa: TRY201 # FIXME CoP

            if (len(ready_metrics) + len(needed_metrics) == 0) or (
                len(ready_metrics) == len(aborted_metrics_info)
            ):
                done = True

        progress_bar.close()  # type: ignore[union-attr]  # Incorrect flagging of 'Item "None" of "Optional[Any]" has no attribute "close"' in external package.

        return aborted_metrics_info

    def _parse(
        self,
        metrics: Dict[MetricConfigurationID, MetricValue],
    ) -> Tuple[Set[MetricConfiguration], Set[MetricConfiguration]]:
        """Given validation graph, returns the ready and needed metrics necessary for validation using a traversal of
        validation graph (a graph structure of metric ids) edges"""  # noqa: E501 # FIXME CoP
        unmet_dependency_ids = set()
        unmet_dependency = set()
        maybe_ready_ids = set()
        maybe_ready = set()

        for edge in self.edges:
            if edge.left.id not in metrics:
                if edge.right is None or edge.right.id in metrics:
                    if edge.left.id not in maybe_ready_ids:
                        maybe_ready_ids.add(edge.left.id)
                        maybe_ready.add(edge.left)
                else:  # noqa: PLR5501 # FIXME CoP
                    if edge.left.id not in unmet_dependency_ids:
                        unmet_dependency_ids.add(edge.left.id)
                        unmet_dependency.add(edge.left)

        return maybe_ready - unmet_dependency, unmet_dependency

    @staticmethod
    def _set_default_metric_kwargs_if_absent(
        default_kwarg_values: dict,
        metric_kwargs: IDDict,
        keys: Tuple[str, ...],
    ) -> None:
        key: str
        for key in keys:
            if (
                key not in metric_kwargs
                and key in default_kwarg_values
                and default_kwarg_values[key] is not None
            ):
                metric_kwargs[key] = default_kwarg_values[key]

    def __repr__(self):  # type: ignore[explicit-override] # FIXME
        edge: MetricEdge
        return ", ".join([edge.__repr__() for edge in self._edges])


class ExpectationValidationGraph:
    def __init__(
        self,
        configuration: ExpectationConfiguration,
        graph: ValidationGraph,
    ) -> None:
        if configuration is None:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                """Instantiation of "ExpectationValidationGraph" requires valid "ExpectationConfiguration" object."""  # noqa: E501 # FIXME CoP
            )

        if graph is None:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                """Instantiation of "ExpectationValidationGraph" requires valid "ValidationGraph" object."""  # noqa: E501 # FIXME CoP
            )

        self._configuration = configuration
        self._graph = graph

    @property
    def configuration(self) -> ExpectationConfiguration:
        return self._configuration

    @property
    def graph(self) -> ValidationGraph:
        return self._graph

    def update(self, graph: ValidationGraph) -> None:
        edge: MetricEdge
        for edge in graph.edges:
            self.graph.add(edge=edge)

    def get_exception_info(
        self,
        metric_info: _AbortedMetricsInfoDict,
    ) -> Dict[str, Union[MetricConfiguration, ExceptionInfo, int]]:
        metric_info = self._filter_metric_info_in_graph(metric_info=metric_info)
        metric_exception_info: Dict[str, Union[MetricConfiguration, ExceptionInfo, int]] = {}
        metric_id: MetricConfigurationID
        metric_info_item: MetricsCalculatorErrorResultValue
        for metric_id, metric_info_item in metric_info.items():
            metric_exception_info[str(metric_id)] = metric_info_item["exception_info"]

        return metric_exception_info

    def _filter_metric_info_in_graph(
        self,
        metric_info: _AbortedMetricsInfoDict,
    ) -> _AbortedMetricsInfoDict:
        graph_metric_ids: List[MetricConfigurationID] = []
        edge: MetricEdge
        vertex: MetricConfiguration
        for edge in self.graph.edges:
            for vertex in [edge.left, edge.right]:
                if vertex is not None:
                    graph_metric_ids.append(vertex.id)

        metric_id: MetricConfigurationID
        metric_info_item: Dict[str, Union[MetricConfiguration, Set[ExceptionInfo], int]]
        return {
            metric_id: metric_info_item
            for metric_id, metric_info_item in metric_info.items()
            if metric_id in graph_metric_ids
        }

    @override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ExpectationValidationGraph):
            return False
        return self.configuration == other.configuration and self.graph == other.graph

    @override
    def __hash__(self) -> int:
        return hash((self.configuration, self.graph))



from __future__ import annotations

import errno
import os
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Dict, Optional, Type, cast

from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.config_substitutor import _ConfigurationSubstitutor
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context.types.base import GXCloudConfig  # noqa: TC001 # FIXME CoP

yaml = YAMLHandler()


class _AbstractConfigurationProvider(ABC):
    def __init__(self) -> None:
        self._substitutor = _ConfigurationSubstitutor()

    @abstractmethod
    def get_values(self) -> Dict[str, str]:
        """
        Retrieve any configuration variables relevant to the provider's environment.
        """
        pass

    def substitute_config(self, config: Any, config_values: Optional[Dict[str, str]] = None) -> Any:
        """
        Utilizes the underlying ConfigurationSubstitutor instance to substitute any
        $VARIABLES with their corresponding config variable value.

        Args:
            config: The config object to update.
            config_values: The dictionary of values to use during the substitution process.
                           If omitted, any values derived from registered providers will be used.

        Returns:
            The input config object with any $VARIABLES replaced with their corresponding config values.
        """  # noqa: E501 # FIXME CoP
        if config_values is None:
            config_values = self.get_values()
        return self._substitutor.substitute_all_config_variables(config, config_values)


class _ConfigurationProvider(_AbstractConfigurationProvider):
    """
    Wrapper class around the other environment-specific configuraiton provider classes.

    Based on relevance, specific providers are registered to this object and are invoked
    using the API defined by the AbstractConfigurationProvider.

    In short, this class' purpose is to aggregate all configuration variables that may
    be present for a given user environment (config variables, env vars, runtime environment, etc.)
    """

    def __init__(self) -> None:
        self._providers: OrderedDict[
            Type[_AbstractConfigurationProvider], _AbstractConfigurationProvider
        ] = OrderedDict()
        super().__init__()

    def register_provider(self, provider: _AbstractConfigurationProvider) -> None:
        """
        Saves a configuration provider to the object's state for downstream usage.
        See `get_values()` for more information.

        Args:
            provider: An instance of a provider to register.
        """
        type_ = type(provider)
        if type_ in self._providers:
            raise ValueError(f"Provider of type {type_} has already been registered!")  # noqa: TRY003 # FIXME CoP
        self._providers[type_] = provider

    def get_provider(
        self, type_: Type[_AbstractConfigurationProvider]
    ) -> Optional[_AbstractConfigurationProvider]:
        """
        Retrieves a registered configuration provider (if available).

        Args:
            type_: The class of the configuration provider to retrieve.

        Returns:
            A registered provider if available.
            If not, None is returned.
        """
        return self._providers.get(type_)

    @override
    def get_values(self) -> Dict[str, str]:
        """
        Iterates through all registered providers to aggregate a list of configuration values.

        Values are generated based on the order of registration; if there is a conflict,
        subsequent providers will overwrite existing values.
        """
        values: Dict[str, str] = {}
        for provider in self._providers.values():
            values.update(provider.get_values())
        return values


class _RuntimeEnvironmentConfigurationProvider(_AbstractConfigurationProvider):
    """
    Responsible for the management of the runtime_environment dictionary provided at runtime.
    """

    def __init__(self, runtime_environment: Dict[str, str]) -> None:
        self._runtime_environment = runtime_environment
        super().__init__()

    @override
    def get_values(self) -> Dict[str, str]:
        return self._runtime_environment


class _EnvironmentConfigurationProvider(_AbstractConfigurationProvider):
    """
    Responsible for the management of environment variables.
    """

    def __init__(self) -> None:
        super().__init__()

    @override
    def get_values(self) -> Dict[str, str]:
        return dict(os.environ)  # noqa: TID251 # os.environ allowed in config files


class _ConfigurationVariablesConfigurationProvider(_AbstractConfigurationProvider):
    """
    Responsible for the management of user-defined configuration variables.

    These can be found in the user's /uncommitted/config_variables.yml file.
    """

    def __init__(
        self, config_variables_file_path: str, root_directory: Optional[str] = None
    ) -> None:
        self._config_variables_file_path = config_variables_file_path
        self._root_directory = root_directory
        super().__init__()

    @override
    def get_values(self) -> Dict[str, str]:
        env_vars = dict(os.environ)  # noqa: TID251 # os.environ allowed in config files
        try:
            # If the user specifies the config variable path with an environment variable, we want to substitute it  # noqa: E501 # FIXME CoP
            defined_path: str = self._substitutor.substitute_config_variable(  # type: ignore[assignment] # FIXME CoP
                self._config_variables_file_path, env_vars
            )
            if not os.path.isabs(defined_path):  # noqa: PTH117 # FIXME CoP
                root_directory: str = self._root_directory or os.curdir
            else:
                root_directory = ""

            var_path = os.path.join(root_directory, defined_path)  # noqa: PTH118 # FIXME CoP
            with open(var_path) as config_variables_file:
                contents = config_variables_file.read()

            variables = dict(yaml.load(contents)) or {}
            return cast(
                "Dict[str, str]",
                self._substitutor.substitute_all_config_variables(variables, env_vars),
            )

        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            return {}


class _CloudConfigurationProvider(_AbstractConfigurationProvider):
    """
    Responsible for the management of a user's GX Cloud credentials.

    See `GXCloudConfig` for more information. Note that this is only registered on the primary
    config provider when in a Cloud-backed environment.
    """

    def __init__(self, cloud_config: GXCloudConfig) -> None:
        self._cloud_config = cloud_config

    @override
    def get_values(self) -> Dict[str, str]:
        from great_expectations.data_context.cloud_constants import (
            GXCloudEnvironmentVariable,
        )

        base_url = self._cloud_config.base_url
        access_token = self._cloud_config.access_token
        organization_id = self._cloud_config.organization_id

        cloud_values: Dict[str, str] = {
            GXCloudEnvironmentVariable.BASE_URL: base_url,
            GXCloudEnvironmentVariable.ACCESS_TOKEN: access_token,
        }

        # organization_id is nullable so we conditionally include it in the output
        if organization_id:
            cloud_values.update(
                {
                    GXCloudEnvironmentVariable.ORGANIZATION_ID: organization_id,
                }
            )

        return cloud_values



import pytest

from great_expectations.checkpoint.checkpoint import CheckpointResult
from great_expectations.render.renderer import OpsgenieRenderer


@pytest.mark.unit
def test_OpsgenieRenderer_render(v1_checkpoint_result: CheckpointResult):
    # Act
    renderer = OpsgenieRenderer()
    raw_output = renderer.render(checkpoint_result=v1_checkpoint_result)
    parts = raw_output.split("\n")

    # Assert
    header = parts.pop(0)  # Separately evaluate header due to dynamic content
    assert "Checkpoint:" in header and "Run ID:" in header
    assert parts == [
        "Status: Failed ❌",
        "",
        "Batch Validation Status: Failed ❌",
        "Expectation Suite Name: my_bad_suite",
        "Data Asset Name: my_first_asset",
        "Run ID: __no_run_id__",
        "Batch ID: my_batch",
        "Summary: 3 of 5 expectations were met",
        "",
        "Batch Validation Status: Success 🎉",
        "Expectation Suite Name: my_good_suite",
        "Data Asset Name: __no_data_asset_name__",
        "Run ID: my_run_id",
        "Batch ID: my_other_batch",
        "Summary: 1 of 1 expectations were met",
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Set, Type, Union

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.exceptions.exceptions import InvalidSetTypeError
from great_expectations.expectations.expectation import (
    BatchExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.render import (
    AtomicDiagnosticRendererType,
    LegacyRendererType,
    RenderedAtomicContent,
    RenderedStringTemplateContent,
    renderedAtomicValueSchema,
)
from great_expectations.render.renderer.observed_value_renderer import ObservedValueRenderState
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import substitute_none_for_missing

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.expectations.metrics.util import CaseInsensitiveString
    from great_expectations.render.renderer_configuration import AddParamArgs


EXPECTATION_SHORT_DESCRIPTION = "Expect the columns in a table to match an unordered set."
COLUMN_SET_DESCRIPTION = (
    "The column names, in any order. In SQL datasources, if the column names are "
    "double quoted, for example '\"column_name\"', a case sensitive match is "
    "done. Otherwise a case insensitive match is done."
)
EXACT_MATCH_DESCRIPTION = (
    "If True, the list of columns must exactly match the observed columns. "
    "If False, observed columns must include column_set but additional columns will pass."
)
DATA_QUALITY_ISSUES = [DataQualityIssues.SCHEMA.value]
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.REDSHIFT.value,
]


class ExpectTableColumnsToMatchSet(BatchExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectTableColumnsToMatchSet is a \
    Batch Expectation.

    BatchExpectations are one of the most common types of Expectation.
    They are evaluated for an entire Batch, and answer a semantic question about the Batch itself.

    Args:
        column_set (list of str): {COLUMN_SET_DESCRIPTION}
        exact_match (boolean): \
            {EXACT_MATCH_DESCRIPTION} Default True.

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[8]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	1.00 	2
            1 	2.30 	5
            2 	4.33 	0

    Code Examples:
        Passing Case:
            Input:
                ExpectTableColumnsToMatchSet(
                    column_set=["test"],
                    exact_match=False
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": [
                      "test",
                      "test2"
                    ],
                    "details": {{
                      "mismatched": {{
                        "unexpected": [
                          "test2"
                        ]
                      }}
                    }}
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectTableColumnsToMatchSet(
                    column_set=["test2", "test3"],
                    exact_match=True
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": [
                      "test",
                      "test2"
                    ],
                    "details": {{
                      "mismatched": {{
                        "unexpected": [
                          "test"
                        ],
                        "missing": [
                          "test3"
                        ]
                      }}
                    }}
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    column_set: Union[list, set, SuiteParameterDict, None] = pydantic.Field(
        description=COLUMN_SET_DESCRIPTION
    )
    exact_match: Union[bool, SuiteParameterDict, None] = pydantic.Field(
        default=True, description=EXACT_MATCH_DESCRIPTION
    )

    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "table expectation"],
        "contributors": [
            "@great_expectations",
        ],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    metric_dependencies = ("table.columns",)
    success_keys = (
        "column_set",
        "exact_match",
    )
    args_keys = (
        "column_set",
        "exact_match",
    )

    class Config:
        title = "Expect table columns to match set"

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type[ExpectTableColumnsToMatchSet]) -> None:
            BatchExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column_set", RendererValueType.ARRAY),
            ("exact_match", RendererValueType.BOOLEAN),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if not params.column_set:
            template_str = "Must specify a set or list of columns."
        else:
            array_param_name = "column_set"
            param_prefix = "column_set_"
            renderer_configuration = cls._add_array_params(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )
            column_set_str: str = cls._get_array_string(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )

            exact_match_str = (
                "exactly" if params.exact_match and params.exact_match.value is True else "at least"
            )

            template_str = (
                f"Must have {exact_match_str} these columns (in any order): {column_set_str}"
            )

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        _ = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(configuration.kwargs, ["column_set", "exact_match"])

        if params["column_set"] is None:
            template_str = "Must specify a set or list of columns."

        else:
            # standardize order of the set for output
            params["column_list"] = list(params["column_set"])

            column_list_template_str = ", ".join(
                [f"$column_list_{idx}" for idx in range(len(params["column_list"]))]
            )

            exact_match_str = "exactly" if params["exact_match"] is True else "at least"

            template_str = f"Must have {exact_match_str} these columns (in any order): {column_list_template_str}"  # noqa: E501 # FIXME CoP

            for idx in range(len(params["column_list"])):
                params[f"column_list_{idx!s}"] = params["column_list"][idx]

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]

    @classmethod
    @renderer(renderer_type=AtomicDiagnosticRendererType.OBSERVED_VALUE)
    @override
    def _atomic_diagnostic_observed_value(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> RenderedAtomicContent:
        renderer_configuration: RendererConfiguration = RendererConfiguration(
            configuration=configuration,
            result=result,
            runtime_configuration=runtime_configuration,
        )
        expected_param_prefix = "exp__"
        expected_param_name = "expected_value"
        ov_param_prefix = "ov__"
        ov_param_name = "observed_value"

        renderer_configuration.add_param(
            name=expected_param_name,
            param_type=RendererValueType.ARRAY,
            value=renderer_configuration.kwargs.get("column_set", []),
        )
        renderer_configuration = cls._add_array_params(
            array_param_name=expected_param_name,
            param_prefix=expected_param_prefix,
            renderer_configuration=renderer_configuration,
        )

        renderer_configuration.add_param(
            name=ov_param_name,
            param_type=RendererValueType.ARRAY,
            value=result.get("result", {}).get("observed_value", []) if result else [],
        )
        renderer_configuration = cls._add_array_params(
            array_param_name=ov_param_name,
            param_prefix=ov_param_prefix,
            renderer_configuration=renderer_configuration,
        )

        observed_columns = (
            (name, sch)
            for name, sch in renderer_configuration.params
            if name.startswith(ov_param_prefix)
        )
        expected_columns = (
            (name, sch)
            for name, sch in renderer_configuration.params
            if name.startswith(expected_param_prefix)
        )
        mismatched_columns = {"unexpected": [], "missing": []}
        if (
            "details" in result["result"]
            and "mismatched" in result["result"]["details"]
            and result["result"]["details"]["mismatched"]
        ):
            mismatched_columns.update(result["result"]["details"]["mismatched"])

        template_str_list = []
        for name, schema in observed_columns:
            render_state = (
                ObservedValueRenderState.UNEXPECTED.value
                if schema.value in mismatched_columns["unexpected"]
                else ObservedValueRenderState.EXPECTED.value
            )
            renderer_configuration.params.__dict__[name].render_state = render_state
            template_str_list.append(f"${name}")

        for name, schema in expected_columns:
            if schema.value in mismatched_columns["missing"]:
                renderer_configuration.params.__dict__[
                    name
                ].render_state = ObservedValueRenderState.MISSING.value
                template_str_list.append(f"${name}")

        renderer_configuration.template_str = " ".join(template_str_list)

        value_obj = renderedAtomicValueSchema.load(
            {
                "template": renderer_configuration.template_str,
                "params": renderer_configuration.params.dict(),
                "meta_notes": renderer_configuration.meta_notes,
                "schema": {"type": "com.superconductive.rendered.string"},
            }
        )
        return RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value=value_obj,
            value_type="StringValueType",
        )

    def _validate(
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        from great_expectations.execution_engine import SqlAlchemyExecutionEngine

        if isinstance(execution_engine, SqlAlchemyExecutionEngine):
            return self._validate_sqlalchemy(metrics)

        # Retrieve expected and observed column names
        expected_column_list = self._get_success_kwargs().get("column_set")
        expected_column_set = (
            set(expected_column_list) if expected_column_list is not None else set()
        )
        actual_column_list = metrics.get("table.columns")
        actual_column_set = set(actual_column_list)

        unmatched_actual_column_set = actual_column_set - expected_column_set
        unmatched_expected_column_set = expected_column_set - actual_column_set
        return _validate_result(
            actual_column_set,
            expected_column_set,
            unmatched_actual_column_set,
            unmatched_expected_column_set,
            self._get_success_kwargs().get("exact_match"),
        )

    def _validate_sqlalchemy(self, metrics: Dict):
        # We want to match the expected columns with the actual columns. We first break up the
        # expected columns into 2 sets, the quoted columns which must match exactly and the unquoted
        # columns, which we case insensitive match.
        expected_column_set = set(self._get_success_kwargs().get("column_set"))
        quoted_expected_column_set = set()
        unquoted_expected_column_set = set()
        for col in expected_column_set:
            if col.startswith('"') and col.endswith('"'):
                quoted_expected_column_set.add(col[1:-1])
            else:
                unquoted_expected_column_set.add(col)

        # The actual columns from the db will be unquoted and may be strs or CaseInsensitiveStrings.
        # We normalize the actual_column_list to CaseInsensitiveStrings so we can use set operations
        # going forward.
        actual_column_list = metrics.get("table.columns")
        actual_column_set = _make_case_insensitive_set(actual_column_list)

        # We make copies of the expected and actual column sets and remove items from them as we
        # find matches between the 2 sets.
        unmatched_expected_column_set = expected_column_set.copy()
        unmatched_actual_column_set = actual_column_set.copy()

        # We first match quoted strings. The expected set is a set of strs while the actual set
        # is a set of CaseInsensitiveStrings so we can't use set operations.
        for col in actual_column_set:
            if str(col) in quoted_expected_column_set:
                unmatched_expected_column_set.remove(f'"{col!s}"')
                unmatched_actual_column_set.remove(col)

        # We normalize the unmatched_expected_column_set to CaseInsensitiveStrings
        unmatched_expected_column_set = _make_case_insensitive_set(unmatched_expected_column_set)

        # We now do the unquoted match
        unquoted_expected_column_set = _make_case_insensitive_set(unquoted_expected_column_set)
        unquoted_matches = unquoted_expected_column_set.intersection(unmatched_actual_column_set)

        # We subtract the unquoted matches from the current unmatched sets to finalize them
        unmatched_actual_column_set = unmatched_actual_column_set - unquoted_matches
        unmatched_expected_column_set = unmatched_expected_column_set - unquoted_matches

        return _validate_result(
            actual_column_set,
            expected_column_set,
            unmatched_actual_column_set,
            unmatched_expected_column_set,
            self._get_success_kwargs().get("exact_match"),
        )


def _make_case_insensitive_set(
    str_set: Optional[set[str | CaseInsensitiveString]],
) -> set[CaseInsensitiveString]:
    """
    Transforms a set of strs to CaseInsensitiveStrings.

    Args:
        str_set: A set of strs.

    Returns:
        A set of CaseInsensitiveString.
    """
    from great_expectations.expectations.metrics.util import (
        CaseInsensitiveString,
    )

    if str_set is None:
        return set()

    case_insensitive_strs = set()
    for s in str_set:
        if isinstance(s, CaseInsensitiveString):
            case_insensitive_strs.add(s)
        elif isinstance(s, str):
            case_insensitive_strs.add(CaseInsensitiveString(s))
        else:
            raise InvalidSetTypeError(
                expected_type="str or CaseInsensitiveString", actual_type=str(type(s))
            )
    return case_insensitive_strs


def _validate_result(
    actual_column_set: Set[Union[str, CaseInsensitiveString]],
    expected_column_set: Set[str],
    unmatched_actual_column_set: Set[Union[str, CaseInsensitiveString]],
    unmatched_expected_column_set: Set[Union[str, CaseInsensitiveString]],
    exact_match: bool,
) -> Dict[str, Any]:
    empty_set = set()
    observed_value = sorted([str(col) for col in actual_column_set])

    if ((expected_column_set is None) and (exact_match is not True)) or (
        unmatched_expected_column_set == empty_set and unmatched_actual_column_set == empty_set
    ):
        return {"success": True, "result": {"observed_value": observed_value}}
    else:
        unexpected_list = sorted([str(col) for col in unmatched_actual_column_set])
        missing_list = sorted([str(col) for col in unmatched_expected_column_set])

        mismatched = {}
        if len(unexpected_list) > 0:
            mismatched["unexpected"] = unexpected_list
        if len(missing_list) > 0:
            mismatched["missing"] = missing_list

        result = {
            "observed_value": observed_value,
            "details": {"mismatched": mismatched},
        }

        return_success = {
            "success": True,
            "result": result,
        }
        return_failed = {
            "success": False,
            "result": result,
        }

        if exact_match:
            return return_failed
        else:  # noqa: PLR5501 # FIXME CoP
            # Failed if there are items in the missing list (but OK to have unexpected_list)
            if len(missing_list) > 0:
                return return_failed
            # Passed if there are no items in the missing list
            else:
                return return_success



from __future__ import annotations

import logging

logger = logging.getLogger(__name__)


class MetaPandasDataset:
    def column_map_expectation(self) -> None:
        logger.debug("MetaPandasDataset.column_map_expectation")



from __future__ import annotations

from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.map_metric_provider import (
    ColumnPairMapMetricProvider,
    column_pair_condition_partial,
)


class ColumnPairValuesEqual(ColumnPairMapMetricProvider):
    condition_metric_name = "column_pair_values.equal"
    condition_domain_keys = (
        "batch_id",
        "table",
        "column_A",
        "column_B",
        "row_condition",
        "condition_parser",
        "ignore_row_if",
    )
    condition_value_keys = ()

    # noinspection PyPep8Naming
    @column_pair_condition_partial(engine=PandasExecutionEngine)
    def _pandas(cls, column_A, column_B, **kwargs):
        return column_A == column_B

    # noinspection PyPep8Naming
    @column_pair_condition_partial(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(cls, column_A, column_B, **kwargs):
        row_wise_cond = sa.and_(
            column_A == column_B,
            sa.not_(sa.or_(column_A == None, column_B == None)),  # noqa: E711 # FIXME CoP
        )
        return row_wise_cond

    # noinspection PyPep8Naming
    @column_pair_condition_partial(engine=SparkDFExecutionEngine)
    def _spark(cls, column_A, column_B, **kwargs):
        row_wise_cond = column_A.eqNullSafe(column_B)
        return row_wise_cond



from __future__ import annotations

import abc
import datetime
import enum
from typing import Callable, ClassVar, List, Type

import ruamel
from dateutil.parser import parse
from ruamel.yaml import yaml_object

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility.typing_extensions import override

yaml = ruamel.yaml.YAML()


@yaml_object(yaml)
class DatePart(enum.Enum):
    """SQL supported date parts for most dialects."""

    YEAR = "year"
    MONTH = "month"
    WEEK = "week"
    DAY = "day"
    HOUR = "hour"
    MINUTE = "minute"
    SECOND = "second"

    @override
    def __eq__(self, other: str | DatePart):  # type: ignore[override] # expects `object`
        if isinstance(other, str):
            return self.value.lower() == other.lower()
        return self.value.lower() == other.value.lower()

    @override
    def __hash__(self: DatePart):
        return hash(self.value)

    @classmethod
    def to_yaml(cls, representer, node):
        """
        Method allows for yaml-encodable representation of ENUM, using internal methods of ruamel.
        pattern was found in the following stackoverflow thread:
        https://stackoverflow.com/questions/48017317/can-ruamel-yaml-encode-an-enum
        """
        return representer.represent_str(data=node.value)


class PartitionerMethod(enum.Enum):
    """The names of available partitioner_methods."""

    PARTITION_ON_YEAR = "partition_on_year"
    PARTITION_ON_YEAR_AND_MONTH = "partition_on_year_and_month"
    PARTITION_ON_YEAR_AND_MONTH_AND_DAY = "partition_on_year_and_month_and_day"
    PARTITION_ON_DATE_PARTS = "partition_on_date_parts"
    PARTITION_ON_WHOLE_TABLE = "partition_on_whole_table"
    PARTITION_ON_COLUMN_VALUE = "partition_on_column_value"
    PARTITION_ON_CONVERTED_DATETIME = "partition_on_converted_datetime"
    PARTITION_ON_DIVIDED_INTEGER = "partition_on_divided_integer"
    PARTITION_ON_MOD_INTEGER = "partition_on_mod_integer"
    PARTITION_ON_MULTI_COLUMN_VALUES = "partition_on_multi_column_values"
    PARTITION_ON_HASHED_COLUMN = "partition_on_hashed_column"

    @override
    def __eq__(self, other: str | PartitionerMethod):  # type: ignore[override] # expects `object`
        if isinstance(other, str):
            return self.value.lower() == other.lower()
        return self.value.lower() == other.value.lower()

    @override
    def __hash__(self: PartitionerMethod):
        return hash(self.value)


class DataPartitioner(abc.ABC):  # noqa: B024 # FIXME CoP
    """Abstract base class containing methods for partitioning data accessible via Execution Engines.

    Note, for convenience, you can also access DatePart via the instance variable
    date_part e.g. DataPartitioner.date_part.MONTH
    """  # noqa: E501 # FIXME CoP

    date_part: ClassVar[Type[DatePart]] = DatePart

    def get_partitioner_method(self, partitioner_method_name: str) -> Callable:
        """Get the appropriate partitioner method from the method name.

        Args:
            partitioner_method_name: name of the partitioner to retrieve.

        Returns:
            partitioner method.
        """
        partitioner_method_name = self._get_partitioner_method_name(partitioner_method_name)

        return getattr(self, partitioner_method_name)

    @staticmethod
    def _get_partitioner_method_name(partitioner_method_name: str) -> str:
        """Accept partitioner methods with or without starting with `_`.

        Args:
            partitioner_method_name: partitioner name starting with or without preceding `_`.

        Returns:
            partitioner method name stripped of preceding underscore.
        """
        if partitioner_method_name.startswith("_"):
            return partitioner_method_name[1:]
        else:
            return partitioner_method_name

    @staticmethod
    def _convert_date_parts(date_parts: List[DatePart] | List[str]) -> List[DatePart]:
        """Convert a list of date parts to DatePart objects.

        Args:
            date_parts: List of DatePart or string representations of DatePart.

        Returns:
            List of DatePart objects
        """
        return [
            DatePart(date_part.lower()) if isinstance(date_part, str) else date_part
            for date_part in date_parts
        ]

    @staticmethod
    def _validate_date_parts(date_parts: List[DatePart] | List[str]) -> None:
        """Validate that date parts exist and are of the correct type.

        Args:
            date_parts: DatePart instances or str.

        Returns:
            None, this method raises exceptions if the config is invalid.
        """
        if len(date_parts) == 0:
            raise gx_exceptions.InvalidConfigError(  # noqa: TRY003 # FIXME CoP
                "date_parts are required when using partition_on_date_parts."
            )
        if not all(isinstance(dp, (DatePart, str)) for dp in date_parts):
            raise gx_exceptions.InvalidConfigError("date_parts should be of type DatePart or str.")  # noqa: TRY003 # FIXME CoP

    @staticmethod
    def _verify_all_strings_are_valid_date_parts(date_part_strings: List[str]) -> None:
        """Verify date part strings by trying to load as DatePart instances.

        Args:
            date_part_strings: A list of strings that should correspond to DatePart.

        Returns:
            None, raises an exception if unable to convert.
        """
        try:
            [DatePart(date_part_string) for date_part_string in date_part_strings]
        except ValueError as e:
            raise gx_exceptions.InvalidConfigError(  # noqa: TRY003 # FIXME CoP
                f"{e} please only specify strings that are supported in DatePart: {[dp.value for dp in DatePart]}"  # noqa: E501 # FIXME CoP
            )

    def _convert_datetime_batch_identifiers_to_date_parts_dict(
        self,
        column_batch_identifiers: datetime.datetime | str | dict,
        date_parts: List[DatePart],
    ) -> dict:
        """Convert batch identifiers to a dict of {date_part as str: date_part value}.

        Args:
            column_batch_identifiers: Batch identifiers related to the column of interest.
            date_parts: List of DatePart to include in the return value.

        Returns:
            A dict of {date_part as str: date_part value} eg. {"day": 3}.
        """

        if isinstance(column_batch_identifiers, str):
            column_batch_identifiers = parse(column_batch_identifiers)

        if isinstance(column_batch_identifiers, datetime.datetime):
            return {
                date_part.value: getattr(column_batch_identifiers, date_part.value)
                for date_part in date_parts
            }
        else:
            self._verify_all_strings_are_valid_date_parts(list(column_batch_identifiers.keys()))
            return column_batch_identifiers



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional, Union

from marshmallow import Schema, fields, post_load

import great_expectations.exceptions as gx_exceptions
from great_expectations._docs_decorators import public_api
from great_expectations.core.data_context_key import DataContextKey
from great_expectations.core.id_dict import BatchKwargs, IDDict
from great_expectations.core.run_identifier import RunIdentifier, RunIdentifierSchema

if TYPE_CHECKING:
    from great_expectations.data_context.cloud_constants import GXCloudRESTResource

logger = logging.getLogger(__name__)


class ExpectationSuiteIdentifier(DataContextKey):
    def __init__(self, name: str) -> None:
        super().__init__()
        if not isinstance(name, str):
            raise gx_exceptions.InvalidDataContextKeyError(  # noqa: TRY003 # FIXME CoP
                f"name must be a string, not {type(name).__name__}"
            )
        self._name = name

    @property
    def name(self):
        return self._name

    def to_tuple(self):  # type: ignore[explicit-override] # FIXME
        return tuple(self.name.split("."))

    def to_fixed_length_tuple(self):  # type: ignore[explicit-override] # FIXME
        return (self.name,)

    @classmethod
    def from_tuple(cls, tuple_):
        return cls(".".join(tuple_))

    @classmethod
    def from_fixed_length_tuple(cls, tuple_):
        return cls(name=tuple_[0])

    def __repr__(self):  # type: ignore[explicit-override] # FIXME
        return f"{self.__class__.__name__}::{self._name}"


class ExpectationSuiteIdentifierSchema(Schema):
    name = fields.Str()

    # noinspection PyUnusedLocal
    @post_load
    def make_expectation_suite_identifier(self, data, **kwargs):
        return ExpectationSuiteIdentifier(**data)


class BatchIdentifier(DataContextKey):
    """A BatchIdentifier tracks"""

    def __init__(
        self,
        batch_identifier: Union[BatchKwargs, dict, str],
        data_asset_name: Optional[str] = None,
    ) -> None:
        super().__init__()
        # if isinstance(batch_identifier, (BatchKwargs, dict)):
        #     self._batch_identifier = batch_identifier.batch_fingerprint

        self._batch_identifier = batch_identifier
        self._data_asset_name = data_asset_name

    @property
    def batch_identifier(self):
        return self._batch_identifier

    @property
    def data_asset_name(self):
        return self._data_asset_name

    def to_tuple(self):  # type: ignore[explicit-override] # FIXME
        return (self.batch_identifier,)

    @classmethod
    def from_tuple(cls, tuple_):
        return cls(batch_identifier=tuple_[0])


class BatchIdentifierSchema(Schema):
    batch_identifier = fields.Str()
    data_asset_name = fields.Str()

    # noinspection PyUnusedLocal
    @post_load
    def make_batch_identifier(self, data, **kwargs):
        return BatchIdentifier(**data)


@public_api
class ValidationResultIdentifier(DataContextKey):
    """A ValidationResultIdentifier identifies a validation result by the fully-qualified expectation_suite_identifier and run_id."""  # noqa: E501 # FIXME CoP

    def __init__(self, expectation_suite_identifier, run_id, batch_identifier) -> None:
        """Constructs a ValidationResultIdentifier

        Args:
            expectation_suite_identifier (ExpectationSuiteIdentifier, list, tuple, or dict):
                identifying information for the fully-qualified expectation suite used to validate
            run_id (RunIdentifier): The run_id for which validation occurred
        """
        super().__init__()
        self._expectation_suite_identifier = expectation_suite_identifier
        if isinstance(run_id, dict):
            run_id = RunIdentifier(**run_id)
        elif run_id is None:
            run_id = RunIdentifier()
        elif not isinstance(run_id, RunIdentifier):
            run_id = RunIdentifier(run_name=str(run_id))

        self._run_id = run_id
        self._batch_identifier = batch_identifier

    @property
    def expectation_suite_identifier(self) -> ExpectationSuiteIdentifier:
        return self._expectation_suite_identifier

    @property
    def run_id(self):
        return self._run_id

    @property
    def batch_identifier(self):
        return self._batch_identifier

    def to_tuple(self):  # type: ignore[explicit-override] # FIXME
        return tuple(
            list(self.expectation_suite_identifier.to_tuple())
            + list(self.run_id.to_tuple())
            + [self.batch_identifier or "__none__"]
        )

    def to_fixed_length_tuple(self):  # type: ignore[explicit-override] # FIXME
        return tuple(
            [self.expectation_suite_identifier.name]
            + list(self.run_id.to_tuple())
            + [self.batch_identifier or "__none__"]
        )

    @classmethod
    def from_tuple(cls, tuple_):
        return cls(
            ExpectationSuiteIdentifier.from_tuple(tuple_[0:-3]),
            RunIdentifier.from_tuple((tuple_[-3], tuple_[-2])),
            tuple_[-1],
        )

    @classmethod
    def from_fixed_length_tuple(cls, tuple_):
        return cls(
            ExpectationSuiteIdentifier(tuple_[0]),
            RunIdentifier.from_tuple((tuple_[1], tuple_[2])),
            tuple_[3],
        )

    @classmethod
    def from_object(cls, validation_result):
        batch_kwargs = validation_result.meta.get("batch_kwargs", {})
        if isinstance(batch_kwargs, IDDict):
            batch_identifier = batch_kwargs.to_id()
        elif isinstance(batch_kwargs, dict):
            batch_identifier = IDDict(batch_kwargs).to_id()
        else:
            raise gx_exceptions.DataContextError(  # noqa: TRY003 # FIXME CoP
                "Unable to construct ValidationResultIdentifier from provided object."
            )
        return cls(
            expectation_suite_identifier=ExpectationSuiteIdentifier(
                validation_result.meta["expectation_suite_name"]
            ),
            run_id=validation_result.meta.get("run_id"),
            batch_identifier=batch_identifier,
        )


class MetricIdentifier(DataContextKey):
    """A MetricIdentifier serves as a key to store and retrieve Metrics."""

    def __init__(self, metric_name, metric_kwargs_id) -> None:
        self._metric_name = metric_name
        self._metric_kwargs_id = metric_kwargs_id

    @property
    def metric_name(self):
        return self._metric_name

    @property
    def metric_kwargs_id(self):
        return self._metric_kwargs_id

    def to_fixed_length_tuple(self):  # type: ignore[explicit-override] # FIXME
        return self.to_tuple()

    def to_tuple(self):  # type: ignore[explicit-override] # FIXME
        if self._metric_kwargs_id is None:
            tuple_metric_kwargs_id = "__"
        else:
            tuple_metric_kwargs_id = self._metric_kwargs_id
        return tuple(
            (self.metric_name, tuple_metric_kwargs_id)
        )  # We use the placeholder in to_tuple

    @classmethod
    def from_fixed_length_tuple(cls, tuple_):
        return cls.from_tuple(tuple_)

    @classmethod
    def from_tuple(cls, tuple_):
        if tuple_[-1] == "__":
            return cls(*tuple_[:-1], None)
        return cls(*tuple_)


class ValidationMetricIdentifier(MetricIdentifier):
    def __init__(
        self,
        run_id,
        data_asset_name,
        expectation_suite_identifier,
        metric_name,
        metric_kwargs_id,
    ) -> None:
        super().__init__(metric_name, metric_kwargs_id)
        if not isinstance(expectation_suite_identifier, ExpectationSuiteIdentifier):
            expectation_suite_identifier = ExpectationSuiteIdentifier(
                name=expectation_suite_identifier
            )

        if isinstance(run_id, dict):
            run_id = RunIdentifier(**run_id)
        elif run_id is None:
            run_id = RunIdentifier()
        elif not isinstance(run_id, RunIdentifier):
            run_id = RunIdentifier(run_name=str(run_id))

        self._run_id = run_id
        self._data_asset_name = data_asset_name
        self._expectation_suite_identifier = expectation_suite_identifier

    @property
    def run_id(self):
        return self._run_id

    @property
    def data_asset_name(self):
        return self._data_asset_name

    @property
    def expectation_suite_identifier(self):
        return self._expectation_suite_identifier

    def to_tuple(self):  # type: ignore[explicit-override] # FIXME
        if self.data_asset_name is None:
            tuple_data_asset_name = "__"
        else:
            tuple_data_asset_name = self.data_asset_name
        return tuple(
            list(self.run_id.to_tuple())
            + [tuple_data_asset_name]
            + list(self.expectation_suite_identifier.to_tuple())
            + [self.metric_name, self.metric_kwargs_id or "__"]
        )

    def to_fixed_length_tuple(self):  # type: ignore[explicit-override] # FIXME
        if self.data_asset_name is None:
            tuple_data_asset_name = "__"
        else:
            tuple_data_asset_name = self.data_asset_name
        return tuple(
            list(self.run_id.to_tuple())
            + [tuple_data_asset_name]
            + list(self.expectation_suite_identifier.to_fixed_length_tuple())
            + [self.metric_name, self.metric_kwargs_id or "__"]
        )

    @classmethod
    def from_tuple(cls, tuple_):
        if len(tuple_) < 6:  # noqa: PLR2004 # FIXME CoP
            raise gx_exceptions.GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                "ValidationMetricIdentifier tuple must have at least six components."
            )
        if tuple_[2] == "__":
            tuple_data_asset_name = None
        else:
            tuple_data_asset_name = tuple_[2]
        metric_id = MetricIdentifier.from_tuple(tuple_[-2:])
        return cls(
            run_id=RunIdentifier.from_tuple((tuple_[0], tuple_[1])),
            data_asset_name=tuple_data_asset_name,
            expectation_suite_identifier=ExpectationSuiteIdentifier.from_tuple(tuple_[3:-2]),
            metric_name=metric_id.metric_name,
            metric_kwargs_id=metric_id.metric_kwargs_id,
        )

    @classmethod
    def from_fixed_length_tuple(cls, tuple_):
        if len(tuple_) != 6:  # noqa: PLR2004 # FIXME CoP
            raise gx_exceptions.GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                "ValidationMetricIdentifier fixed length tuple must have exactly six components."
            )
        if tuple_[2] == "__":
            tuple_data_asset_name = None
        else:
            tuple_data_asset_name = tuple_[2]
        metric_id = MetricIdentifier.from_tuple(tuple_[-2:])
        return cls(
            run_id=RunIdentifier.from_fixed_length_tuple((tuple_[0], tuple_[1])),
            data_asset_name=tuple_data_asset_name,
            expectation_suite_identifier=ExpectationSuiteIdentifier.from_fixed_length_tuple(
                tuple((tuple_[3],))
            ),
            metric_name=metric_id.metric_name,
            metric_kwargs_id=metric_id.metric_kwargs_id,
        )


class GXCloudIdentifier(DataContextKey):
    def __init__(
        self,
        resource_type: GXCloudRESTResource,
        id: str | None = None,
        resource_name: str | None = None,
    ) -> None:
        super().__init__()

        self._resource_type = resource_type
        self._id = id
        self._resource_name = resource_name

    @property
    def resource_type(self) -> GXCloudRESTResource:
        return self._resource_type

    @resource_type.setter
    def resource_type(self, value: GXCloudRESTResource) -> None:
        self._resource_type = value

    @property
    def id(self) -> str | None:
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        self._id = value

    @property
    def resource_name(self) -> str | None:
        return self._resource_name

    def to_tuple(self):  # type: ignore[explicit-override] # FIXME
        return (self.resource_type, self.id, self.resource_name)

    def to_fixed_length_tuple(self):  # type: ignore[explicit-override] # FIXME
        return self.to_tuple()

    @classmethod
    def from_tuple(cls, tuple_):
        # Only add resource name if it exists in the tuple_
        if len(tuple_) == 3:  # noqa: PLR2004 # FIXME CoP
            return cls(resource_type=tuple_[0], id=tuple_[1], resource_name=tuple_[2])
        return cls(resource_type=tuple_[0], id=tuple_[1])

    @classmethod
    def from_fixed_length_tuple(cls, tuple_):
        return cls.from_tuple(tuple_)

    def __repr__(self):  # type: ignore[explicit-override] # FIXME
        repr = f"{self.__class__.__name__}::{self.resource_type}::{self.id}"
        if self.resource_name:
            repr += f"::{self.resource_name}"
        return repr


class ValidationResultIdentifierSchema(Schema):
    expectation_suite_identifier = fields.Nested(
        ExpectationSuiteIdentifierSchema,
        required=True,
        error_messages={
            "required": "expectation_suite_identifier is required for a ValidationResultIdentifier"
        },
    )
    run_id = fields.Nested(
        RunIdentifierSchema,
        required=True,
        error_messages={"required": "run_id is required for a ValidationResultIdentifier"},
    )
    batch_identifier = fields.Nested(BatchIdentifierSchema, required=True)

    # noinspection PyUnusedLocal
    @post_load
    def make_validation_result_identifier(self, data, **kwargs):
        return ValidationResultIdentifier(**data)


class SiteSectionIdentifier(DataContextKey):
    def __init__(self, site_section_name, resource_identifier) -> None:
        self._site_section_name = site_section_name
        if site_section_name in ["validations", "profiling"]:
            if isinstance(resource_identifier, ValidationResultIdentifier):
                self._resource_identifier = resource_identifier
            elif isinstance(resource_identifier, (tuple, list)):
                self._resource_identifier = ValidationResultIdentifier(*resource_identifier)
            else:
                self._resource_identifier = ValidationResultIdentifier(**resource_identifier)
        elif site_section_name == "expectations":
            if isinstance(resource_identifier, ExpectationSuiteIdentifier):
                self._resource_identifier = resource_identifier  # type: ignore[assignment] # FIXME CoP
            elif isinstance(resource_identifier, (tuple, list)):
                self._resource_identifier = ExpectationSuiteIdentifier(  # type: ignore[assignment] # FIXME CoP
                    *resource_identifier
                )
            else:
                self._resource_identifier = ExpectationSuiteIdentifier(  # type: ignore[assignment] # FIXME CoP
                    **resource_identifier
                )
        else:
            raise gx_exceptions.InvalidDataContextKeyError(  # noqa: TRY003 # FIXME CoP
                "SiteSectionIdentifier only supports 'validations' and 'expectations' as site section names"  # noqa: E501 # FIXME CoP
            )

    @property
    def site_section_name(self):
        return self._site_section_name

    @property
    def resource_identifier(self):
        return self._resource_identifier

    def to_tuple(self):  # type: ignore[explicit-override] # FIXME
        site_section_identifier_tuple_list = [self.site_section_name] + list(
            self.resource_identifier.to_tuple()
        )
        return tuple(site_section_identifier_tuple_list)

    @classmethod
    def from_tuple(cls, tuple_):
        if tuple_[0] == "validations":
            return cls(
                site_section_name=tuple_[0],
                resource_identifier=ValidationResultIdentifier.from_tuple(tuple_[1:]),
            )
        elif tuple_[0] == "expectations":
            return cls(
                site_section_name=tuple_[0],
                resource_identifier=ExpectationSuiteIdentifier.from_tuple(tuple_[1:]),
            )
        else:
            raise gx_exceptions.InvalidDataContextKeyError(  # noqa: TRY003 # FIXME CoP
                "SiteSectionIdentifier only supports 'validations' and 'expectations' as site section names"  # noqa: E501 # FIXME CoP
            )


class ConfigurationIdentifier(DataContextKey):
    def __init__(self, configuration_key: str) -> None:
        super().__init__()
        if not isinstance(configuration_key, str):
            raise gx_exceptions.InvalidDataContextKeyError(  # noqa: TRY003 # FIXME CoP
                f"configuration_key must be a string, not {type(configuration_key).__name__}"
            )
        self._configuration_key = configuration_key

    @property
    def configuration_key(self) -> str:
        return self._configuration_key

    def to_tuple(self):  # type: ignore[explicit-override] # FIXME
        return tuple(self.configuration_key.split("."))

    def to_fixed_length_tuple(self):  # type: ignore[explicit-override] # FIXME
        return (self.configuration_key,)

    @classmethod
    def from_tuple(cls, tuple_):
        return cls(".".join(tuple_))

    @classmethod
    def from_fixed_length_tuple(cls, tuple_):
        return cls(configuration_key=tuple_[0])

    def __repr__(self):  # type: ignore[explicit-override] # FIXME
        return f"{self.__class__.__name__}::{self._configuration_key}"


class ConfigurationIdentifierSchema(Schema):
    configuration_key = fields.Str()

    # noinspection PyUnusedLocal
    @post_load
    def make_configuration_identifier(self, data, **kwargs):
        return ConfigurationIdentifier(**data)


expectationSuiteIdentifierSchema = ExpectationSuiteIdentifierSchema()
validationResultIdentifierSchema = ValidationResultIdentifierSchema()
runIdentifierSchema = RunIdentifierSchema()
batchIdentifierSchema = BatchIdentifierSchema()
configurationIdentifierSchema = ConfigurationIdentifierSchema()



from __future__ import annotations

import copy
from typing import Any, Dict, List
from unittest import mock

import pytest

from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations import registry
from great_expectations.expectations.metrics.map_metric_provider import (
    ColumnMapMetricProvider,
    ColumnPairMapMetricProvider,
    MulticolumnMapMetricProvider,
)
from great_expectations.expectations.metrics.map_metric_provider.column_condition_partial import (
    column_condition_partial,
)
from great_expectations.expectations.metrics.map_metric_provider.column_pair_condition_partial import (  # noqa: E501 # FIXME CoP
    column_pair_condition_partial,
)
from great_expectations.expectations.metrics.map_metric_provider.multicolumn_condition_partial import (  # noqa: E501 # FIXME CoP
    multicolumn_condition_partial,
)
from great_expectations.expectations.metrics.metric_provider import (
    MetricProvider,
    metric_value,
)
from great_expectations.expectations.metrics.query_metric_provider import (
    QueryMetricProvider,
    QueryParameters,
)
from great_expectations.expectations.metrics.table_metric_provider import (
    TableMetricProvider,
)
from great_expectations.expectations.metrics.util import MAX_RESULT_RECORDS
from tests.expectations.metrics.conftest import MockResult, MockSqlAlchemyExecutionEngine

pytestmark = pytest.mark.unit


@pytest.fixture
def mock_registry(monkeypatch: pytest.MonkeyPatch):
    """Ensures consistent `_registered_metrics` state among test cases runs."""
    monkeypatch.setattr(
        registry,
        "_registered_metrics",
        copy.deepcopy(registry._registered_metrics),
        raising=True,
    )
    yield registry


def test__base_metric_provider__registration(mock_registry):
    """This tests whether the MetricProvider class registers the correct metrics."""
    registered_metric_keys = list(mock_registry._registered_metrics.keys())
    for key in registered_metric_keys:
        assert "custom_metric" not in key

    prev_registered_metric_key_count = len(registered_metric_keys)

    class CustomMetricProvider(MetricProvider):
        metric_name = "custom_metric"
        value_keys = ()

        @metric_value(engine=PandasExecutionEngine)
        def _pandas(
            cls,
            execution_engine: PandasExecutionEngine,
            metric_domain_kwargs: dict,
            metric_value_kwargs: dict,
            metrics: Dict[str, Any],
            runtime_configuration: dict,
        ):
            raise NotImplementedError

        @metric_value(engine=SqlAlchemyExecutionEngine)
        def _sqlalchemy(
            cls,
            execution_engine: SqlAlchemyExecutionEngine,
            metric_domain_kwargs: dict,
            metric_value_kwargs: dict,
            metrics: Dict[str, Any],
            runtime_configuration: dict,
        ):
            raise NotImplementedError

        @metric_value(engine=SparkDFExecutionEngine)
        def _spark(
            cls,
            execution_engine: SparkDFExecutionEngine,
            metric_domain_kwargs: dict,
            metric_value_kwargs: dict,
            metrics: Dict[str, Any],
            runtime_configuration: dict,
        ):
            raise NotImplementedError

    CustomMetricProvider()

    assert len(mock_registry._registered_metrics.keys()) == prev_registered_metric_key_count + 1
    assert "custom_metric" in mock_registry._registered_metrics


def test__table_metric_provider__registration(mock_registry):
    """This tests whether the TableMetricProvider class registers the correct metrics."""
    registered_metric_keys = list(mock_registry._registered_metrics.keys())
    for key in registered_metric_keys:
        assert "table.custom_metric" not in key

    prev_registered_metric_key_count = len(registered_metric_keys)

    class CustomTableMetricProvider(TableMetricProvider):
        metric_name = "table.custom_metric"

        @metric_value(engine=PandasExecutionEngine)
        def _pandas(
            cls,
            execution_engine: PandasExecutionEngine,
            metric_domain_kwargs: dict,
            metric_value_kwargs: dict,
            metrics: Dict[str, Any],
            runtime_configuration: dict,
        ):
            raise NotImplementedError

        @metric_value(engine=SqlAlchemyExecutionEngine)
        def _sqlalchemy(
            cls,
            execution_engine: SqlAlchemyExecutionEngine,
            metric_domain_kwargs: dict,
            metric_value_kwargs: dict,
            metrics: Dict[str, Any],
            runtime_configuration: dict,
        ):
            raise NotImplementedError

        @metric_value(engine=SparkDFExecutionEngine)
        def _spark(
            cls,
            execution_engine: SparkDFExecutionEngine,
            metric_domain_kwargs: dict,
            metric_value_kwargs: dict,
            metrics: Dict[str, Any],
            runtime_configuration: dict,
        ):
            raise NotImplementedError

    CustomTableMetricProvider()

    assert len(mock_registry._registered_metrics.keys()) == prev_registered_metric_key_count + 1
    assert "table.custom_metric" in mock_registry._registered_metrics


def test__column_map_metric__registration(mock_registry):
    """This tests whether the ColumnMapMetricProvider class registers the correct metrics.

    The actual logic for this lives in the private method: `_register_metric_functions`, which is invoked from within `__new__` for the ancestor class `MetricProvider`.

    Since _register_metric_functions is private, we don't want to test it directly. Instead, we declare a custom ColumnMapMetricProvider, and test that the correct metrics are registered.
    """  # noqa: E501 # FIXME CoP
    registered_metric_keys = list(mock_registry._registered_metrics.keys())
    for key in registered_metric_keys:
        assert "column_values.equal_seven" not in key

    prev_registered_metric_key_count = len(registered_metric_keys)

    class CustomColumnValuesEqualSeven(ColumnMapMetricProvider):
        condition_metric_name = "column_values.equal_seven"

        @column_condition_partial(engine=PandasExecutionEngine)
        def _pandas(cls, column, **kwargs):
            return column == 7

        @column_condition_partial(engine=SqlAlchemyExecutionEngine)
        def _sqlalchemy(cls, column, **kwargs):
            # return column.in_([3])
            return column.is_(7)

        @column_condition_partial(engine=SparkDFExecutionEngine)
        def _spark(cls, column, **kwargs):
            return column.contains(7)

    CustomColumnValuesEqualSeven()

    assert len(mock_registry._registered_metrics.keys()) == prev_registered_metric_key_count + 8

    new_keys = [
        "column_values.equal_seven.condition",
        "column_values.equal_seven.unexpected_count",
        "column_values.equal_seven.unexpected_index_list",
        "column_values.equal_seven.unexpected_index_query",
        "column_values.equal_seven.unexpected_rows",
        "column_values.equal_seven.unexpected_values",
        "column_values.equal_seven.unexpected_value_counts",
        "column_values.equal_seven.unexpected_count.aggregate_fn",
    ]
    for key in new_keys:
        assert key in mock_registry._registered_metrics


def test__column_pair_map_metric__registration(mock_registry):
    """This tests whether the ColumnPairMapMetricProvider class registers the correct metrics."""
    registered_metric_keys = list(mock_registry._registered_metrics.keys())
    for key in registered_metric_keys:
        assert "column_pair_values.equal_seven" not in key

    prev_registered_metric_key_count = len(registered_metric_keys)

    class CustomColumnPairValuesEqualSeven(ColumnPairMapMetricProvider):
        condition_metric_name = "column_pair_values.equal_seven"

        @column_pair_condition_partial(engine=PandasExecutionEngine)
        def _pandas(cls, column_A, column_B, **kwargs):
            raise NotImplementedError

        @column_pair_condition_partial(engine=SqlAlchemyExecutionEngine)
        def _sqlalchemy(cls, column_A, column_B, _dialect, **kwargs):
            raise NotImplementedError

        @column_pair_condition_partial(engine=SparkDFExecutionEngine)
        def _spark(cls, column_A, column_B, **kwargs):
            raise NotImplementedError

    CustomColumnPairValuesEqualSeven()

    assert len(mock_registry._registered_metrics.keys()) == prev_registered_metric_key_count + 7

    for key in mock_registry._registered_metrics:
        if "column_pair_values.equal_seven" in key:
            print(key)

    new_keys = [
        "column_pair_values.equal_seven.condition",
        "column_pair_values.equal_seven.unexpected_count",
        "column_pair_values.equal_seven.unexpected_index_list",
        "column_pair_values.equal_seven.unexpected_index_query",
        "column_pair_values.equal_seven.unexpected_rows",
        "column_pair_values.equal_seven.unexpected_values",
        "column_pair_values.equal_seven.filtered_row_count",
    ]
    for key in new_keys:
        assert key in mock_registry._registered_metrics


def test__multicolumn_map_metric__registration(mock_registry):
    """This tests whether the MultiColumnMapMetricProvider class registers the correct metrics."""
    registered_metric_keys = list(mock_registry._registered_metrics.keys())
    for key in registered_metric_keys:
        assert "multicolumn_values.equal_seven" not in key

    prev_registered_metric_key_count = len(registered_metric_keys)

    class CustomMultiColumnValuesEqualSeven(MulticolumnMapMetricProvider):
        condition_metric_name = "multicolumn_values.equal_seven"

        condition_domain_keys = (
            "batch_id",
            "table",
            "column_list",
            "row_condition",
            "condition_parser",
            "ignore_row_if",
        )
        condition_value_keys = ()

        @multicolumn_condition_partial(engine=PandasExecutionEngine)
        def _pandas(cls, column_list, **kwargs):
            raise NotImplementedError

        @multicolumn_condition_partial(engine=SqlAlchemyExecutionEngine)
        def _sqlalchemy(cls, column_list, **kwargs):
            raise NotImplementedError

        @multicolumn_condition_partial(engine=SparkDFExecutionEngine)
        def _spark(cls, column_list, **kwargs):
            raise NotImplementedError

    CustomMultiColumnValuesEqualSeven()

    assert len(mock_registry._registered_metrics.keys()) == prev_registered_metric_key_count + 7

    new_keys = [
        "multicolumn_values.equal_seven.condition",
        "multicolumn_values.equal_seven.unexpected_count",
        "multicolumn_values.equal_seven.unexpected_index_list",
        "multicolumn_values.equal_seven.unexpected_index_query",
        "multicolumn_values.equal_seven.unexpected_rows",
        "multicolumn_values.equal_seven.unexpected_values",
        "multicolumn_values.equal_seven.filtered_row_count",
    ]
    for key in new_keys:
        assert key in mock_registry._registered_metrics


def test__query_metric_provider__registration(mock_registry):
    """This tests whether the QueryMetricProvider class registers the correct metrics."""
    registered_metric_keys = list(mock_registry._registered_metrics.keys())
    for key in registered_metric_keys:
        assert "query.custom_metric" not in key

    prev_registered_metric_key_count = len(registered_metric_keys)

    class CustomQueryMetricProvider(QueryMetricProvider):
        metric_name = "query.custom_metric"

        @metric_value(engine=SqlAlchemyExecutionEngine)
        def _sqlalchemy(
            cls,
            execution_engine: SqlAlchemyExecutionEngine,
            metric_domain_kwargs: dict,
            metric_value_kwargs: dict,
            metrics: Dict[str, Any],
            runtime_configuration: dict,
        ) -> List[dict]:
            raise NotImplementedError

        @metric_value(engine=SparkDFExecutionEngine)
        def _spark(
            cls,
            execution_engine: SparkDFExecutionEngine,
            metric_domain_kwargs: dict,
            metric_value_kwargs: dict,
            metrics: Dict[str, Any],
            runtime_configuration: dict,
        ) -> List[dict]:
            raise NotImplementedError

    CustomQueryMetricProvider()

    assert len(mock_registry._registered_metrics.keys()) == prev_registered_metric_key_count + 1
    assert "query.custom_metric" in mock_registry._registered_metrics


@pytest.mark.unit
@pytest.mark.parametrize(
    "query_parameters,expected_dict",
    [
        (
            None,
            {},
        ),
        (
            QueryParameters(),
            {},
        ),
        (
            QueryParameters(column="my_column"),
            {"column": "my_column"},
        ),
        (
            QueryParameters(column_A="my_column_A", column_B="my_column_B"),
            {"column_A": "my_column_A", "column_B": "my_column_B"},
        ),
        (
            QueryParameters(columns=["my_column_A", "my_column_B", "my_column_C"]),
            {"col_1": "my_column_A", "col_2": "my_column_B", "col_3": "my_column_C"},
        ),
    ],
)
def test__get_parameters_dict_from_query_parameters(
    query_parameters: QueryParameters, expected_dict: dict
):
    actual_dict = QueryMetricProvider._get_parameters_dict_from_query_parameters(query_parameters)
    assert actual_dict == expected_dict


@pytest.mark.unit
@mock.patch.object(sa, "text")
@pytest.mark.parametrize(
    "batch_selectable,expected_query",
    [
        (
            sa.table("my_table"),
            "SELECT my_column FROM (my_table) WHERE passenger_count > 7",
        ),
        (
            sa.select("*").select_from(sa.text("my_table")).subquery(),
            "SELECT my_column FROM (SELECT * \nFROM my_table) AS subselect "
            "WHERE passenger_count > 7",
        ),
        (
            sa.select("*").select_from(sa.text("my_table")),
            "SELECT my_column FROM (SELECT * \nFROM my_table) AS subselect "
            "WHERE passenger_count > 7",
        ),
    ],
)
def test_get_sqlalchemy_records_from_query_and_batch_selectable__query(
    mock_sqlalchemy_text,
    mock_sqlalchemy_execution_engine: MockSqlAlchemyExecutionEngine,
    batch_selectable: sa.Selectable,
    expected_query: str,
):
    mock_sqlalchemy_text.return_value = "*"
    with mock.patch.object(mock_sqlalchemy_execution_engine, "execute_query"):
        substituted_batch_subquery = (
            QueryMetricProvider._get_substituted_batch_subquery_from_query_and_batch_selectable(
                query="SELECT {column} FROM {batch} WHERE passenger_count > 7",
                batch_selectable=batch_selectable,
                execution_engine=mock_sqlalchemy_execution_engine,
                query_parameters=QueryParameters(column="my_column"),
            )
        )
        QueryMetricProvider._get_sqlalchemy_records_from_substituted_batch_subquery(
            substituted_batch_subquery=substituted_batch_subquery,
            execution_engine=mock_sqlalchemy_execution_engine,
        )
    assert substituted_batch_subquery == expected_query
    mock_sqlalchemy_text.assert_called_with(expected_query)


@pytest.mark.unit
@mock.patch.object(MockResult, "fetchmany")
def test_get_sqlalchemy_records_from_query_and_batch_selectable__record_count(
    mock_sqlalchemy_fetchmany,
    mock_sqlalchemy_execution_engine: MockSqlAlchemyExecutionEngine,
):
    mock_sqlalchemy_fetchmany.return_value = []
    substituted_batch_subquery = (
        QueryMetricProvider._get_substituted_batch_subquery_from_query_and_batch_selectable(
            query="SELECT * FROM {batch} WHERE passenger_count > 7",
            batch_selectable=sa.select("*").select_from(sa.text("my_table")).subquery(),
            execution_engine=mock_sqlalchemy_execution_engine,
            query_parameters=QueryParameters(column="my_column"),
        )
    )
    QueryMetricProvider._get_sqlalchemy_records_from_substituted_batch_subquery(
        substituted_batch_subquery=substituted_batch_subquery,
        execution_engine=mock_sqlalchemy_execution_engine,
    )
    mock_sqlalchemy_fetchmany.assert_called_with(MAX_RESULT_RECORDS)



from __future__ import annotations

import logging
import pathlib
import shutil
import sys
import uuid
import warnings
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    Final,
    Generator,
    Literal,
    Mapping,
    Protocol,
    Sequence,
    TypedDict,
)

import pytest
from packaging.version import Version
from pytest import param

import great_expectations.expectations.core as gxe
from great_expectations.checkpoint.checkpoint import Checkpoint
from great_expectations.compatibility.sqlalchemy import (
    DatabaseError as SqlAlchemyDatabaseError,
)
from great_expectations.compatibility.sqlalchemy import (
    OperationalError as SqlAlchemyOperationalError,
)
from great_expectations.compatibility.sqlalchemy import (
    ProgrammingError as SqlAlchemyProgrammingError,
)
from great_expectations.compatibility.sqlalchemy import (
    TextClause,
    engine,
    inspect,
    quoted_name,
)
from great_expectations.compatibility.sqlalchemy import (
    __version__ as sqlalchemy_version,
)
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context import EphemeralDataContext
from great_expectations.datasource.fluent import (
    PostgresDatasource,
    SQLDatasource,
    SqliteDatasource,
)
from great_expectations.execution_engine.sqlalchemy_dialect import (
    DIALECT_IDENTIFIER_QUOTE_STRINGS,
    GXSqlDialect,
    quote_str,
)
from great_expectations.expectations.expectation_configuration import ExpectationConfiguration
from tests.integration.fluent.conftest import TEST_TABLE_NAME

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from great_expectations.checkpoint.checkpoint import CheckpointResult
    from great_expectations.execution_engine import SqlAlchemyExecutionEngine

TERMINAL_WIDTH: Final = shutil.get_terminal_size().columns
STAR_SEPARATOR: Final = "*" * TERMINAL_WIDTH

PYTHON_VERSION: Final[Literal["py38", "py39", "py310", "py311"]] = (
    f"py{sys.version_info.major}{sys.version_info.minor}"  # type: ignore[assignment] # str for each python version
)
SQLA_VERSION: Final = Version(sqlalchemy_version or "0.0.0")
LOGGER: Final = logging.getLogger("tests")

TRINO_TABLE: Final[str] = "customer"

# NOTE: can we create tables in trino?
# some of the trino tests probably don't make sense if we can't create tables
DO_NOT_CREATE_TABLES: set[str] = {"trino"}
# sqlite db files should be using fresh tmp_path on every test
DO_NOT_DROP_TABLES: set[str] = {"sqlite"}

DatabaseType: TypeAlias = Literal["postgres", "sqlite", "trino"]
TableNameCase: TypeAlias = Literal[
    "quoted_lower",
    "quoted_mixed",
    "quoted_upper",
    "unquoted_lower",
    "unquoted_mixed",
    "unquoted_upper",
]

# TODO: simplify this and possible get rid of this mapping once we have settled on
# all the naming conventions we want to support for different SQL dialects
# NOTE: commented out are tests we know fail for individual datasources. Ideally all
# test cases should work for all datasrouces
TABLE_NAME_MAPPING: Final[Mapping[DatabaseType, Mapping[TableNameCase, str]]] = {
    "postgres": {
        "unquoted_lower": TEST_TABLE_NAME.lower(),
        "quoted_lower": f'"{TEST_TABLE_NAME.lower()}"',
        # "unquoted_upper": TEST_TABLE_NAME.upper(),
        "quoted_upper": f'"{TEST_TABLE_NAME.upper()}"',
        "quoted_mixed": f'"{TEST_TABLE_NAME.title()}"',
        # "unquoted_mixed": TEST_TABLE_NAME.title(),
    },
    "trino": {
        "unquoted_lower": TRINO_TABLE.lower(),
        "quoted_lower": f"'{TRINO_TABLE.lower()}'",
        # "unquoted_upper": TRINO_TABLE.upper(),
        # "quoted_upper": f"'{TRINO_TABLE.upper()}'",
        # "quoted_mixed": f"'TRINO_TABLE.title()'",
        # "unquoted_mixed": TRINO_TABLE.title(),
    },
    "sqlite": {
        "unquoted_lower": TEST_TABLE_NAME.lower(),
        "quoted_lower": f'"{TEST_TABLE_NAME.lower()}"',
        "unquoted_upper": TEST_TABLE_NAME.upper(),
        "quoted_upper": f'"{TEST_TABLE_NAME.upper()}"',
        "quoted_mixed": f'"{TEST_TABLE_NAME.title()}"',
        "unquoted_mixed": TEST_TABLE_NAME.title(),
    },
}

# column names
UNQUOTED_UPPER_COL: Final = "UNQUOTED_UPPER_COL"
UNQUOTED_LOWER_COL: Final = "unquoted_lower_col"
QUOTED_UPPER_COL: Final = "QUOTED_UPPER_COL"
QUOTED_LOWER_COL: Final = "quoted_lower_col"
QUOTED_MIXED_CASE: Final = "quotedMixed"
QUOTED_W_DOTS: Final = "quoted.w.dots"


def get_random_identifier_name() -> str:
    guid = uuid.uuid4()
    return f"i{guid.hex}"


RAND_SCHEMA: Final[str] = f"{PYTHON_VERSION}_{get_random_identifier_name()}"


def _get_exception_details(
    result: CheckpointResult,
    prettyprint: bool = False,
) -> list[
    dict[
        Literal["exception_message", "exception_traceback", "raised_exception"],
        str,
    ]
]:
    """Extract a list of exception_info dicts from a CheckpointResult."""
    first_run_result = next(iter(result.run_results.values()))
    validation_results = first_run_result.results
    if prettyprint:
        print(f"validation_result.results:\n{pf(validation_results, depth=2)}\n")

    exc_details = [
        r["exception_info"]
        for r in validation_results
        if r["exception_info"].get("raised_exception")
    ]
    if exc_details and prettyprint:
        print(f"{len(exc_details)} exception_info(s):\n{STAR_SEPARATOR}")
        for i, exc_info in enumerate(exc_details, start=1):
            print(
                f"  {i}: {exc_info['exception_message']}"
                f"\n\n{exc_info['exception_traceback']}\n{STAR_SEPARATOR}"
            )
    return exc_details


@pytest.fixture
def capture_engine_logs(caplog: pytest.LogCaptureFixture) -> pytest.LogCaptureFixture:
    """Capture SQLAlchemy engine logs and display them if the test fails."""
    caplog.set_level(logging.INFO, logger="sqlalchemy.engine")
    return caplog


class Row(TypedDict):
    id: int
    name: str
    quoted_upper_col: str
    quoted_lower_col: str
    quoted_mixed_case: str
    quoted_w_dots: str
    unquoted_upper_col: str
    unquoted_lower_col: str


ColNameParams: TypeAlias = Literal[
    # DDL: unquoted_lower_col ------
    "unquoted_lower_col",
    '"unquoted_lower_col"',
    "UNQUOTED_LOWER_COL",
    '"UNQUOTED_LOWER_COL"',
    # DDL: UNQUOTED_UPPER_COL ------
    "unquoted_upper_col",
    '"unquoted_upper_col"',
    "UNQUOTED_UPPER_COL",
    '"UNQUOTED_UPPER_COL"',
    # DDL: "quoted_lower_col" -----
    "quoted_lower_col",
    '"quoted_lower_col"',
    "QUOTED_LOWER_COL",
    '"QUOTED_LOWER_COL"',
    # DDL: "QUOTED_UPPER_COL" ----
    "quoted_upper_col",
    '"quoted_upper_col"',
    "QUOTED_UPPER_COL",
    '"QUOTED_UPPER_COL"',
    # DDL: "quotedMixed" -----
    "quotedmixed",
    "quotedMixed",
    '"quotedMixed"',
    "QUOTEDMIXED",
    # DDL: "quoted.w.dots" -------
    "quoted.w.dots",
    '"quoted.w.dots"',
    "QUOTED.W.DOTS",
    '"QUOTED.W.DOTS"',
]

ColNameParamId: TypeAlias = Literal[
    # DDL: unquoted_lower_col ------
    "str unquoted_lower_col",
    'str "unquoted_lower_col"',
    "str UNQUOTED_LOWER_COL",
    'str "UNQUOTED_LOWER_COL"',
    # DDL: UNQUOTED_UPPER_COL ------
    "str unquoted_upper_col",
    'str "unquoted_upper_col"',
    "str UNQUOTED_UPPER_COL",
    'str "UNQUOTED_UPPER_COL"',
    # DDL: "quoted_lower_col" -----
    "str quoted_lower_col",
    'str "quoted_lower_col"',
    "str QUOTED_LOWER_COL",
    'str "QUOTED_LOWER_COL"',
    # DDl: "QUOTED_UPPER_COL" ----
    "str quoted_upper_col",
    'str "quoted_upper_col"',
    "str QUOTED_UPPER_COL",
    'str "QUOTED_UPPER_COL"',
    # DDL: "quotedMixed" -----
    "str quotedmixed",
    "str quotedMixed",
    'str "quotedMixed"',
    "str QUOTEDMIXED",
    # DDL: "quoted.w.dots" -------
    "str quoted.w.dots",
    'str "quoted.w.dots"',
    "str QUOTED.W.DOTS",
    'str "QUOTED.W.DOTS"',
]

COLUMN_DDL: Final[Mapping[ColNameParams, str]] = {
    # DDL: unquoted_lower_col ------
    "unquoted_lower_col": UNQUOTED_LOWER_COL,
    '"unquoted_lower_col"': UNQUOTED_LOWER_COL,
    "UNQUOTED_LOWER_COL": UNQUOTED_LOWER_COL,
    '"UNQUOTED_LOWER_COL"': UNQUOTED_LOWER_COL,
    # DDL: UNQUOTED_UPPER_COL ------
    "unquoted_upper_col": UNQUOTED_UPPER_COL,
    '"unquoted_upper_col"': UNQUOTED_UPPER_COL,
    "UNQUOTED_UPPER_COL": UNQUOTED_UPPER_COL,
    '"UNQUOTED_UPPER_COL"': UNQUOTED_UPPER_COL,
    # DDL: "quoted_lower_col" -----
    "quoted_lower_col": f'"{QUOTED_LOWER_COL}"',
    '"quoted_lower_col"': f'"{QUOTED_LOWER_COL}"',
    "QUOTED_LOWER_COL": f'"{QUOTED_LOWER_COL}"',
    '"QUOTED_LOWER_COL"': f'"{QUOTED_LOWER_COL}"',
    # DDl: "QUOTED_UPPER_COL" ----
    "quoted_upper_col": f'"{QUOTED_UPPER_COL}"',
    '"quoted_upper_col"': f'"{QUOTED_UPPER_COL}"',
    "QUOTED_UPPER_COL": f'"{QUOTED_UPPER_COL}"',
    '"QUOTED_UPPER_COL"': f'"{QUOTED_UPPER_COL}"',
    # DDL: "quotedMixed" -----
    "quotedmixed": f"{QUOTED_MIXED_CASE.lower()}",
    "quotedMixed": f"{QUOTED_MIXED_CASE}",
    '"quotedMixed"': f'"{QUOTED_MIXED_CASE}"',
    "QUOTEDMIXED": f"{QUOTED_MIXED_CASE.upper()}",
    # DDL: "quoted.w.dots" -------
    "quoted.w.dots": f'"{QUOTED_W_DOTS}"',
    '"quoted.w.dots"': f'"{QUOTED_W_DOTS}"',
    "QUOTED.W.DOTS": f'"{QUOTED_W_DOTS}"',
    '"QUOTED.W.DOTS"': f'"{QUOTED_W_DOTS}"',
}


# TODO: remove items from this lookup when working on fixes
# The presence of a database type in the list indicates that this specific value fails when
# used as the `column_name` value for at least one expectation.
# It does not mean that it SHOULD or SHOULD NOT fail, but that it currently does.
FAILS_EXPECTATION: Final[Mapping[ColNameParamId, list[DatabaseType]]] = {
    # DDL: unquoted_lower_col ------
    "str unquoted_lower_col": [],
    'str "unquoted_lower_col"': ["postgres", "sqlite"],
    "str UNQUOTED_LOWER_COL": ["postgres", "sqlite"],
    'str "UNQUOTED_LOWER_COL"': ["sqlite"],
    # DDL: UNQUOTED_UPPER_COL ------
    "str unquoted_upper_col": ["sqlite"],
    'str "unquoted_upper_col"': ["postgres", "sqlite"],
    "str UNQUOTED_UPPER_COL": ["postgres"],
    'str "UNQUOTED_UPPER_COL"': ["postgres", "sqlite"],
    # DDL: "quoted_lower_col" -----
    'str "quoted_lower_col"': ["postgres", "sqlite"],
    "str QUOTED_LOWER_COL": ["postgres", "sqlite"],
    'str "QUOTED_LOWER_COL"': ["sqlite"],
    # DDl: "QUOTED_UPPER_COL" ----
    "str quoted_upper_col": ["sqlite", "postgres"],
    'str "quoted_upper_col"': ["sqlite"],
    "str QUOTED_UPPER_COL": [],
    'str "QUOTED_UPPER_COL"': ["postgres", "sqlite"],
    # DDL: "quotedMixed" -----
    "str quotedmixed": [
        "postgres",
        "sqlite",
    ],
    'str "quotedMixed"': [
        "postgres",
        "sqlite",
    ],
    "str QUOTEDMIXED": [
        "postgres",
        "sqlite",
    ],
    # DDL: "quoted.w.dots" -------
    'str "quoted.w.dots"': ["postgres", "sqlite"],
    "str QUOTED.W.DOTS": ["sqlite", "postgres"],
    'str "QUOTED.W.DOTS"': ["sqlite"],
}


class TableFactory(Protocol):
    def __call__(
        self,
        gx_engine: SqlAlchemyExecutionEngine,
        table_names: set[str],
        schema: str | None = None,
        data: Sequence[Row] = ...,
    ) -> None: ...


@pytest.fixture(
    scope="class",
)
def table_factory() -> Generator[TableFactory, None, None]:  # noqa: C901 # FIXME CoP
    """
    Class scoped.
    Given a SQLALchemy engine, table_name and schema,
    create the table if it does not exist and drop it after the test class.
    """
    all_created_tables: dict[str, list[dict[Literal["table_name", "schema"], str | None]]] = {}
    engines: dict[str, engine.Engine] = {}

    def _table_factory(
        gx_engine: SqlAlchemyExecutionEngine,
        table_names: set[str],
        schema: str | None = None,
        data: Sequence[Row] = tuple(),
    ) -> None:
        sa_engine = gx_engine.engine
        if sa_engine.dialect.name in DO_NOT_CREATE_TABLES:
            LOGGER.info(f"Skipping table creation for {table_names} for {sa_engine.dialect.name}")
            return
        LOGGER.info(
            f"SQLA:{SQLA_VERSION} - Creating `{sa_engine.dialect.name}` table for {table_names} if it does not exist"  # noqa: E501 # FIXME CoP
        )
        dialect = GXSqlDialect(sa_engine.dialect.name)
        created_tables: list[dict[Literal["table_name", "schema"], str | None]] = []

        with gx_engine.get_connection() as conn:
            quoted_upper_col: str = quote_str(QUOTED_UPPER_COL, dialect=dialect)
            quoted_lower_col: str = quote_str(QUOTED_LOWER_COL, dialect=dialect)
            quoted_w_dots: str = quote_str(QUOTED_W_DOTS, dialect=dialect)
            quoted_mixed_case: str = quote_str(QUOTED_MIXED_CASE, dialect=dialect)
            transaction = conn.begin()
            if schema:
                conn.execute(TextClause(f"CREATE SCHEMA IF NOT EXISTS {schema}"))
            for name in table_names:
                qualified_table_name = f"{schema}.{name}" if schema else name
                # TODO: use dialect specific quotes
                create_tables: str = (
                    f"CREATE TABLE IF NOT EXISTS {qualified_table_name}"
                    " (id INTEGER, name VARCHAR(255),"
                    f" {quoted_upper_col} VARCHAR(255), {quoted_lower_col} VARCHAR(255),"
                    f" {UNQUOTED_UPPER_COL} VARCHAR(255), {UNQUOTED_LOWER_COL} VARCHAR(255),"
                    f" {quoted_mixed_case} VARCHAR(255), {quoted_w_dots} VARCHAR(255))"
                )
                conn.execute(TextClause(create_tables))
                if data:
                    insert_data = (
                        f"INSERT INTO {qualified_table_name}"
                        f" (id, name, {quoted_upper_col}, {quoted_lower_col},"
                        f" {UNQUOTED_UPPER_COL}, {UNQUOTED_LOWER_COL},"
                        f" {quoted_mixed_case}, {quoted_w_dots})"
                        " VALUES (:id, :name, :quoted_upper_col, :quoted_lower_col,"
                        " :unquoted_upper_col, :unquoted_lower_col,"
                        " :quoted_mixed_case, :quoted_w_dots)"
                    )
                    conn.execute(TextClause(insert_data), data)

                created_tables.append(dict(table_name=name, schema=schema))
            transaction.commit()
        all_created_tables[sa_engine.dialect.name] = created_tables
        engines[sa_engine.dialect.name] = sa_engine

    yield _table_factory

    # teardown
    print(f"dropping tables\n{pf(all_created_tables)}")
    for dialect, tables in all_created_tables.items():
        if dialect in DO_NOT_DROP_TABLES:
            print(f"skipping drop for {dialect}")
            continue
        engine = engines[dialect]
        with engine.connect() as conn:
            transaction = conn.begin()
            schema: str | None = None
            for table in tables:
                name = table["table_name"]
                schema = table["schema"]
                qualified_table_name = f"{schema}.{name}" if schema else name
                conn.execute(TextClause(f"DROP TABLE IF EXISTS {qualified_table_name}"))
            if schema:
                conn.execute(TextClause(f"DROP SCHEMA IF EXISTS {schema}"))
            transaction.commit()


@pytest.fixture
def trino_ds(context: EphemeralDataContext) -> SQLDatasource:
    ds = context.data_sources.add_sql(
        "trino",
        connection_string="trino://user:@localhost:8088/tpch/sf1",
    )
    return ds


@pytest.fixture
def postgres_ds(context: EphemeralDataContext) -> PostgresDatasource:
    ds = context.data_sources.add_postgres(
        "postgres",
        connection_string="postgresql+psycopg2://postgres:postgres@localhost:5432/test_ci",
    )
    return ds


@pytest.fixture
def sqlite_ds(context: EphemeralDataContext, tmp_path: pathlib.Path) -> SqliteDatasource:
    ds = context.data_sources.add_sqlite(
        "sqlite", connection_string=f"sqlite:///{tmp_path}/test.db"
    )
    return ds


@pytest.fixture(
    params=[
        param(
            "trino",
            marks=[
                pytest.mark.trino,
                pytest.mark.skip(reason="cannot create trino tables"),
            ],
        ),
        param("postgres", marks=[pytest.mark.postgresql]),
        param("sqlite", marks=[pytest.mark.sqlite]),
    ]
)
def self_hosted_sql_datasources(
    request: pytest.FixtureRequest,
    capture_engine_logs: pytest.LogCaptureFixture,
) -> Generator[SQLDatasource, None, None]:
    datasource = request.getfixturevalue(f"{request.param}_ds")
    yield datasource


@pytest.mark.parametrize(
    "asset_name",
    [
        param("unquoted_lower"),
        param("quoted_lower"),
        param("unquoted_upper"),
        param("quoted_upper"),
        param("quoted_mixed"),
        param("unquoted_mixed"),
    ],
)
class TestTableIdentifiers:
    @pytest.mark.trino
    def test_trino(self, trino_ds: SQLDatasource, asset_name: TableNameCase):
        table_name = TABLE_NAME_MAPPING["trino"].get(asset_name)
        if not table_name:
            pytest.skip(f"no '{asset_name}' table_name for trino")

        table_names: list[str] = inspect(trino_ds.get_engine()).get_table_names()
        print(f"trino tables:\n{pf(table_names)}))")

        trino_ds.add_table_asset(asset_name, table_name=table_name)

    @pytest.mark.postgresql
    def test_postgres(
        self,
        postgres_ds: PostgresDatasource,
        asset_name: TableNameCase,
        table_factory: TableFactory,
    ):
        table_name = TABLE_NAME_MAPPING["postgres"].get(asset_name)
        if not table_name:
            pytest.skip(f"no '{asset_name}' table_name for postgres")
        # create table
        table_factory(gx_engine=postgres_ds.get_execution_engine(), table_names={table_name})

        table_names: list[str] = inspect(postgres_ds.get_engine()).get_table_names()
        print(f"postgres tables:\n{pf(table_names)}))")

        postgres_ds.add_table_asset(asset_name, table_name=table_name)

    @pytest.mark.sqlite
    def test_sqlite(
        self,
        sqlite_ds: SqliteDatasource,
        asset_name: TableNameCase,
        table_factory: TableFactory,
    ):
        table_name = TABLE_NAME_MAPPING["sqlite"][asset_name]
        # create table
        table_factory(
            gx_engine=sqlite_ds.get_execution_engine(),
            table_names={table_name},
        )

        table_names: list[str] = inspect(sqlite_ds.get_engine()).get_table_names()
        print(f"sqlite tables:\n{pf(table_names)}))")

        sqlite_ds.add_table_asset(asset_name, table_name=table_name)

    @pytest.mark.filterwarnings(  # snowflake `add_table_asset` raises warning on passing a schema
        "once::great_expectations.datasource.fluent.GxDatasourceWarning"
    )
    @pytest.mark.parametrize(
        "datasource_type,schema",
        [
            param("trino", None, marks=[pytest.mark.trino]),
            param("postgres", None, marks=[pytest.mark.postgresql]),
            param("sqlite", None, marks=[pytest.mark.sqlite]),
        ],
    )
    def test_checkpoint_run(
        self,
        request: pytest.FixtureRequest,
        context: EphemeralDataContext,
        table_factory: TableFactory,
        asset_name: TableNameCase,
        datasource_type: DatabaseType,
        schema: str | None,
    ):
        datasource: SQLDatasource = request.getfixturevalue(f"{datasource_type}_ds")

        table_name: str | None = TABLE_NAME_MAPPING[datasource_type].get(asset_name)
        if not table_name:
            pytest.skip(f"no '{asset_name}' table_name for {datasource_type}")

        # create table
        table_factory(
            gx_engine=datasource.get_execution_engine(),
            table_names={table_name},
            schema=schema,
        )

        with warnings.catch_warnings():
            # passing a schema to snowflake tables is deprecated
            warnings.simplefilter("once", DeprecationWarning)
            asset = datasource.add_table_asset(
                asset_name, table_name=table_name, schema_name=schema
            )
            batch_definition = asset.add_batch_definition_whole_table("whole table!")

        suite = context.suites.add(ExpectationSuite(name=f"{datasource.name}-{asset.name}"))
        suite.add_expectation(gxe.ExpectColumnValuesToNotBeNull(column="name"))

        validation_definition = context.validation_definitions.add(
            ValidationDefinition(name="validation_definition", suite=suite, data=batch_definition)
        )
        checkpoint = context.checkpoints.add(
            Checkpoint(
                name=f"{datasource.name}-{asset.name}",
                validation_definitions=[validation_definition],
            )
        )
        result = checkpoint.run()

        _get_exception_details(result, prettyprint=True)
        assert result.success is True


def _fails_expectation(param_id: str) -> bool:
    """
    Lookup whether a given column_name fails an expectation for a given dialect.
    This does not mean that it SHOULD fail, but that it currently does.
    """
    column_name: ColNameParamId
    dialect, column_name, *_ = param_id.split("-")  # type: ignore[assignment] # FIXME CoP
    dialects_need_fixes: list[DatabaseType] = FAILS_EXPECTATION.get(column_name, [])
    return dialect in dialects_need_fixes


def _is_quote_char_dialect_mismatch(
    dialect: GXSqlDialect,
    column_name: str | quoted_name,
) -> bool:
    quote_char = column_name[0] if column_name[0] in ("'", '"', "`") else None
    if quote_char:
        dialect_quote_char = DIALECT_IDENTIFIER_QUOTE_STRINGS[dialect]
        if quote_char != dialect_quote_char:
            return True
    return False


def _raw_query_check_column_exists(
    column_name_param: str,
    qualified_table_name: str,
    gx_execution_engine: SqlAlchemyExecutionEngine,
) -> bool:
    """Use a simple 'SELECT {column_name_param} from {qualified_table_name};' query to check if the column exists.'"""  # noqa: E501 # FIXME CoP
    with gx_execution_engine.get_connection() as connection:
        query = f"""SELECT {column_name_param} FROM {qualified_table_name} LIMIT 1;"""
        print(f"query:\n  {query}")
        # an exception will be raised if the column does not exist
        try:
            col_exist_check = connection.execute(
                TextClause(query),
            )
            print(f"\nResults:\n  {col_exist_check.keys()}")
            col_exist_result = col_exist_check.fetchone()
            print(f"  Values: {col_exist_result}\n{column_name_param} exists!\n")
        except (
            SqlAlchemyProgrammingError,
            SqlAlchemyOperationalError,
            SqlAlchemyDatabaseError,
        ) as sql_err:
            LOGGER.debug("_raw_query_check_column_exists - SQLAlchemy Error", exc_info=sql_err)
            print(f"\n{column_name_param} does not exist! - {sql_err.__class__.__name__}\n")
            return False
        return True


@pytest.mark.filterwarnings(
    "once::DeprecationWarning"
)  # snowflake `add_table_asset` raises warning on passing a schema
class TestColumnExpectations:
    @pytest.mark.parametrize(
        "column_name",
        [
            # DDL: unquoted_lower_col ----------------------------------
            param("unquoted_lower_col", id="str unquoted_lower_col"),
            param("UNQUOTED_LOWER_COL", id="str UNQUOTED_LOWER_COL"),
            # DDL: UNQUOTED_UPPER_COL ----------------------------------
            param("unquoted_upper_col", id="str unquoted_upper_col"),
            param("UNQUOTED_UPPER_COL", id="str UNQUOTED_UPPER_COL"),
            # DDL: "quoted_lower_col"-----------------------------------
            param("quoted_lower_col", id="str quoted_lower_col"),
            param("QUOTED_LOWER_COL", id="str QUOTED_LOWER_COL"),
            # DDL: "QUOTED_UPPER_COL" ----------------------------------
            param("quoted_upper_col", id="str quoted_upper_col"),
            param("QUOTED_UPPER_COL", id="str QUOTED_UPPER_COL"),
            # DDL: "quotedMixed" -------------------------------------
            param("quotedmixed", id="str quotedmixed"),
            param("quotedMixed", id="str quotedMixed"),
            param("QUOTEDMIXED", id="str QUOTEDMIXED"),
            # DDL: "quoted.w.dots" -------------------------------------
            param("quoted.w.dots", id="str quoted.w.dots"),
            param("QUOTED.W.DOTS", id="str QUOTED.W.DOTS"),
        ],
    )
    def test_unquoted_params(
        self,
        context: EphemeralDataContext,
        self_hosted_sql_datasources: SQLDatasource,
        table_factory: TableFactory,
        column_name: str | quoted_name,
        request: pytest.FixtureRequest,
    ):
        """
        Test column expectations when using unquoted column name parameters
        (actual column may have DDL with quotes).

        Test fails if the expectation fails regardless of dialect.
        """
        param_id = request.node.callspec.id
        datasource = self_hosted_sql_datasources

        if column_name[0] in ("'", '"', "`"):
            pytest.skip(f"see _desired_state tests for {column_name!r}")
        elif _fails_expectation(param_id):
            # apply marker this way so that xpasses can be seen in the report
            request.applymarker(pytest.mark.xfail(run=False))

        print(f"\ncolumn DDL:\n  {COLUMN_DDL[column_name]}")  # type: ignore[index] # FIXME
        print(f"\n`column_name` parameter __repr__:\n  {column_name!r}")
        print(f"type:\n  {type(column_name)}\n")

        table_factory(
            gx_engine=datasource.get_execution_engine(),
            table_names={TEST_TABLE_NAME},
            data=[
                {
                    "id": 1,
                    "name": param_id,
                    "quoted_upper_col": "my column is uppercase",
                    "quoted_lower_col": "my column is lowercase",
                    "unquoted_upper_col": "whatever",
                    "unquoted_lower_col": "whatever",
                    "quoted_mixed_case": "Whatever",
                    "quoted_w_dots": "what.ever",
                },
            ],
        )

        asset = datasource.add_table_asset("my_asset", table_name=TEST_TABLE_NAME)
        print(f"asset:\n{asset!r}\n")

        suite = context.suites.add(ExpectationSuite(name=f"{datasource.name}-{asset.name}"))
        suite.add_expectation_configuration(
            expectation_configuration=ExpectationConfiguration(
                type="expect_column_values_to_match_regex",
                kwargs={"column": column_name, "regex": r".*"},
            )
        )
        suite.save()

        batch_definition = asset.add_batch_definition_whole_table("my_batch_def")
        validation_definition = context.validation_definitions.add(
            ValidationDefinition(name="my_validation_def", suite=suite, data=batch_definition)
        )

        checkpoint = context.checkpoints.add(
            checkpoint=Checkpoint(
                name=f"{datasource.name}-{asset.name}_checkpoint",
                validation_definitions=[validation_definition],
            )
        )
        result = checkpoint.run()

        _ = _get_exception_details(result, prettyprint=True)

        assert result.success is True, "validation failed"

    @pytest.mark.parametrize(
        "column_name",
        [
            # DDL: unquoted_lower_col ----------------------------------
            param('"unquoted_lower_col"', id='str "unquoted_lower_col"'),
            # DDL: UNQUOTED_UPPER_COL ----------------------------------
            param('"UNQUOTED_UPPER_COL"', id='str "UNQUOTED_UPPER_COL"'),
            # DDL: "quoted_lower_col"-----------------------------------
            param('"quoted_lower_col"', id='str "quoted_lower_col"'),
            # DDL: "QUOTED_UPPER_COL" ----------------------------------
            param('"QUOTED_UPPER_COL"', id='str "QUOTED_UPPER_COL"'),
            # DDL: "quotedMixed" -------------------------------------
            param('"quotedMixed"', id='str "quotedMixed"'),
            # DDL: "quoted.w.dots" -------------------------------------
            param('"quoted.w.dots"', id='str "quoted.w.dots"'),
        ],
    )
    def test_quoted_params(
        self,
        context: EphemeralDataContext,
        self_hosted_sql_datasources: SQLDatasource,
        table_factory: TableFactory,
        column_name: str | quoted_name,
        request: pytest.FixtureRequest,
    ):
        """
        Test column expectations when using quoted column name parameters
        (actual column may have DDL without quotes).

        Test fails if the expectation fails regardless of dialect.
        """
        param_id = request.node.callspec.id
        datasource = self_hosted_sql_datasources
        dialect = GXSqlDialect(datasource.get_engine().dialect.name)

        if column_name[0] not in ("'", '"', "`"):
            pytest.skip(f"see test_unquoted_params for {column_name!r}")
        elif _is_quote_char_dialect_mismatch(dialect, column_name):
            pytest.skip(f"quote char dialect mismatch: {column_name[0]}")
        elif _fails_expectation(param_id):
            # apply marker this way so that xpasses can be seen in the report
            request.applymarker(pytest.mark.xfail(run=False))

        schema: str | None = None

        print(f"\ncolumn DDL:\n  {COLUMN_DDL[column_name]}")  # type: ignore[index] # FIXME
        print(f"\n`column_name` parameter __repr__:\n  {column_name!r}")
        print(f"type:\n  {type(column_name)}\n")

        table_factory(
            gx_engine=datasource.get_execution_engine(),
            table_names={TEST_TABLE_NAME},
            schema=schema,
            data=[
                {
                    "id": 1,
                    "name": param_id,
                    "quoted_upper_col": "my column is uppercase",
                    "quoted_lower_col": "my column is lowercase",
                    "unquoted_upper_col": "whatever",
                    "unquoted_lower_col": "whatever",
                    "quoted_mixed_case": "Whatever",
                    "quoted_w_dots": "what.ever",
                },
            ],
        )

        asset = datasource.add_table_asset(
            "my_asset", table_name=TEST_TABLE_NAME, schema_name=schema
        )
        print(f"asset:\n{asset!r}\n")

        suite = context.suites.add(ExpectationSuite(name=f"{datasource.name}-{asset.name}"))
        suite.add_expectation_configuration(
            expectation_configuration=ExpectationConfiguration(
                type="expect_column_values_to_match_regex",
                kwargs={"column": column_name, "regex": r".*"},
            )
        )
        suite.save()

        batch_definition = asset.add_batch_definition_whole_table("my_batch_def")
        validation_definition = context.validation_definitions.add(
            ValidationDefinition(name="my_validation_def", suite=suite, data=batch_definition)
        )

        checkpoint = context.checkpoints.add(
            checkpoint=Checkpoint(
                name=f"{datasource.name}-{asset.name}_checkpoint",
                validation_definitions=[validation_definition],
            )
        )
        result = checkpoint.run()

        assert result.success is True, "validation failed"

    @pytest.mark.parametrize(
        "column_name",
        [
            # DDL: unquoted_lower_col ----------------------------------
            param("unquoted_lower_col", id="str unquoted_lower_col"),
            param('"unquoted_lower_col"', id='str "unquoted_lower_col"'),
            param("UNQUOTED_LOWER_COL", id="str UNQUOTED_LOWER_COL"),
            param('"UNQUOTED_LOWER_COL"', id='str "UNQUOTED_LOWER_COL"'),
            # DDL: UNQUOTED_UPPER_COL ----------------------------------
            param("unquoted_upper_col", id="str unquoted_upper_col"),
            param('"unquoted_upper_col"', id='str "unquoted_upper_col"'),
            param("UNQUOTED_UPPER_COL", id="str UNQUOTED_UPPER_COL"),
            param('"UNQUOTED_UPPER_COL"', id='str "UNQUOTED_UPPER_COL"'),
            # DDL: "quoted_lower_col"-----------------------------------
            param("quoted_lower_col", id="str quoted_lower_col"),
            param('"quoted_lower_col"', id='str "quoted_lower_col"'),
            param("QUOTED_LOWER_COL", id="str QUOTED_LOWER_COL"),
            param('"QUOTED_LOWER_COL"', id='str "QUOTED_LOWER_COL"'),
            # DDL: "QUOTED_UPPER_COL" ----------------------------------
            param("quoted_upper_col", id="str quoted_upper_col"),
            param('"quoted_upper_col"', id='str "quoted_upper_col"'),
            param("QUOTED_UPPER_COL", id="str QUOTED_UPPER_COL"),
            param('"QUOTED_UPPER_COL"', id='str "QUOTED_UPPER_COL"'),
            # DDL: "quotedMixed" ---------------------------------------
            param("quotedmixed", id="str quotedmixed"),
            param("quotedMixed", id="str quotedMixed"),
            param('"quotedMixed"', id='str "quotedMixed"'),
            param("QUOTEDMIXED", id="str QUOTEDMIXED"),
            # DDL: "quoted.w.dots" -------------------------------------
            param("quoted.w.dots", id="str quoted.w.dots"),
            param('"quoted.w.dots"', id='str "quoted.w.dots"'),  # TODO: fix me
            param("QUOTED.W.DOTS", id="str QUOTED.W.DOTS"),
            param('"QUOTED.W.DOTS"', id='str "QUOTED.W.DOTS"'),
        ],
    )
    def test_desired_state(
        self,
        context: EphemeralDataContext,
        self_hosted_sql_datasources: SQLDatasource,
        table_factory: TableFactory,
        column_name: str | quoted_name,
        request: pytest.FixtureRequest,
    ):
        """
        Perform a raw query to check if the column exists before running the expectation.
        This is used to determine how the identifier behaves natively in the database and
        therefore determine if the expectation should pass or fail.

        An expectation is expected to succeed if the column 'exists' and fail if it does not.
        If we want GX to behave the same way as each dialect/database,
        these tests should be our guide.

        However currently, GX does not behave the same way as the databases in all cases.
        """
        param_id = request.node.callspec.id
        datasource = self_hosted_sql_datasources
        dialect = GXSqlDialect(datasource.get_engine().dialect.name)

        original_column_name = column_name
        if column_name.startswith('"') and column_name.endswith('"'):
            # databricks uses backticks for quoting
            column_name = quote_str(column_name[1:-1], dialect=dialect)

        print(f"\ncolumn DDL:\n  {COLUMN_DDL[original_column_name]}")  # type: ignore[index] # FIXME
        print(f"\n`column_name` parameter __repr__:\n  {column_name!r}")
        print(f"type:\n  {type(column_name)}\n")

        table_factory(
            gx_engine=datasource.get_execution_engine(),
            table_names={TEST_TABLE_NAME},
            data=[
                {
                    "id": 1,
                    "name": param_id,
                    "quoted_upper_col": "my column is uppercase",
                    "quoted_lower_col": "my column is lowercase",
                    "unquoted_upper_col": "whatever",
                    "unquoted_lower_col": "whatever",
                    "quoted_mixed_case": "Whatever",
                    "quoted_w_dots": "what.ever",
                },
            ],
        )

        qualified_table_name: str = TEST_TABLE_NAME
        # check the column exists so that we know what if the expectation should succeed or fail
        column_exists = _raw_query_check_column_exists(
            column_name,
            qualified_table_name,
            datasource.get_execution_engine(),
        )

        asset = datasource.add_table_asset("my_asset", table_name=TEST_TABLE_NAME)
        print(f"asset:\n{asset!r}\n")

        suite = context.suites.add(ExpectationSuite(name=f"{datasource.name}-{asset.name}"))
        suite.add_expectation_configuration(
            expectation_configuration=ExpectationConfiguration(
                type="expect_column_values_to_match_regex",
                kwargs={"column": column_name, "regex": r".*"},
            )
        )
        suite.save()

        batch_definition = asset.add_batch_definition_whole_table("my_batch_def")
        validation_definition = context.validation_definitions.add(
            ValidationDefinition(name="my_validation_def", suite=suite, data=batch_definition)
        )

        checkpoint = context.checkpoints.add(
            checkpoint=Checkpoint(
                name=f"{datasource.name}-{asset.name}_checkpoint",
                validation_definitions=[validation_definition],
            )
        )
        result = checkpoint.run()

        try:
            if column_exists:
                assert result.success is True, "column exists but validation failed"
            else:
                assert result.success is False, "column does not exist but validation succeeded"
        except AssertionError as ae:
            reason = str(ae).splitlines()[0]
            print(reason)
            # xfail if the expectation doesn't behave as the dialect would
            pytest.xfail(reason=reason)


if __name__ == "__main__":
    pytest.main([__file__, "-vv", "-rXf"])



import uuid
from typing import Dict, Optional, Set

import boto3
import pyparsing as pp
import pytest
from moto import mock_s3

from great_expectations.data_context import get_context
from great_expectations.data_context.store import StoreBackend, TupleS3StoreBackend
from great_expectations.data_context.types.base import DataContextConfig


def build_in_code_data_context_project_config(
    bucket: str = "leakybucket",
    expectations_store_prefix: str = "expectations_store_prefix",
    checkpoint_store_prefix: str = "checkpoint_store_prefix",
    validation_results_store_prefix: str = "validation_results_store_prefix",
    data_docs_store_prefix: str = "data_docs_store_prefix",
    stores: Optional[Dict] = None,
) -> DataContextConfig:
    """
    Create a project config for an in-code data context.
    Not a fixture because we want to control when this is built (after the expectation store).
    Args:
        expectations_store_prefix: prefix for expectations store
        checkpoint_store_prefix: prefix for checkpoint store
        validation_results_store_prefix: prefix for validations store
        data_docs_store_prefix: prefix for data docs
        bucket: name of the s3 bucket
        stores: optional overwrite of the default stores

    Returns:
        DataContextConfig using s3 for all stores.
    """
    if stores is None:
        stores = {
            "expectations_S3_store": {
                "class_name": "ExpectationsStore",
                "store_backend": {
                    "class_name": "TupleS3StoreBackend",
                    "bucket": bucket,
                    "prefix": expectations_store_prefix,
                },
            },
            "validation_results_S3_store": {
                "class_name": "ValidationResultsStore",
                "store_backend": {
                    "class_name": "TupleS3StoreBackend",
                    "bucket": bucket,
                    "prefix": validation_results_store_prefix,
                },
            },
            "checkpoint_store": {"class_name": "CheckpointStore"},
        }
    project_config = DataContextConfig(
        config_version=2,
        plugins_directory=None,
        config_variables_file_path=None,
        stores=stores,
        checkpoint_store_name="checkpoint_store",
        expectations_store_name="expectations_S3_store",
        validation_results_store_name="validation_results_S3_store",
        data_docs_sites={
            "s3_site": {
                "class_name": "SiteBuilder",
                "store_backend": {
                    "class_name": "TupleS3StoreBackend",
                    "bucket": bucket,
                    "prefix": data_docs_store_prefix,
                },
                "site_index_builder": {
                    "class_name": "DefaultSiteIndexBuilder",
                },
            }
        },
    )
    return project_config


def get_store_backend_id_from_s3(bucket: str, prefix: str, key: str) -> uuid.UUID:
    """
    Return the UUID store_backend_id from a given s3 file
    Args:
        bucket: s3 bucket
        prefix: prefix for s3 bucket
        key: filename in s3 bucket

    Returns:

    """
    s3_response_object = boto3.client("s3").get_object(Bucket=bucket, Key=f"{prefix}/{key}")
    ge_store_backend_id_file_contents = s3_response_object["Body"].read().decode("utf-8")

    store_backend_id_file_parser = StoreBackend.STORE_BACKEND_ID_PREFIX + pp.Word(pp.hexnums + "-")
    parsed_store_backend_id = store_backend_id_file_parser.parseString(
        ge_store_backend_id_file_contents
    )
    return uuid.UUID(parsed_store_backend_id[1])


def list_s3_bucket_contents(bucket: str, prefix: str) -> Set[str]:
    """
    List the contents of an s3 bucket as a set of strings given bucket name and prefix
    Args:
        bucket: s3 bucket
        prefix: prefix for s3 bucket

    Returns:
        set of filepath strings
    """
    return {
        s3_object_info["Key"]
        for s3_object_info in boto3.client("s3").list_objects_v2(Bucket=bucket, Prefix=prefix)[
            "Contents"
        ]
    }


@pytest.mark.aws_deps
@mock_s3
def test_DataContext_construct_data_context_id_uses_id_of_currently_configured_expectations_store(
    aws_credentials,
):
    """
    What does this test and why?

    A DataContext should have an id. This ID should come from either:
    1. configured expectations store store_backend_id
    2. great_expectations.yml
    3. new generated id from DataContextConfig
    This test verifies that DataContext._construct_data_context_id
    uses the store_backend_id from the currently configured expectations store
    when instantiating the DataContext
    """

    store_backend_id_filename = StoreBackend.STORE_BACKEND_ID_KEY[0]
    bucket = "leakybucket"
    expectations_store_prefix = "expectations_store_prefix"
    validation_results_store_prefix = "validation_results_store_prefix"
    data_docs_store_prefix = "data_docs_store_prefix"
    data_context_prefix = ""

    # Create a bucket in Moto's mock AWS environment
    conn = boto3.resource("s3", region_name="us-east-1")
    conn.create_bucket(Bucket=bucket)

    # Create a TupleS3StoreBackend
    # Initialize without store_backend_id and check that the store_backend_id is generated correctly
    s3_expectations_store_backend = TupleS3StoreBackend(
        filepath_template="my_file_{0}",
        bucket=bucket,
        prefix=expectations_store_prefix,
    )
    # Make sure store_backend_id is not the error string
    store_error_uuid = uuid.UUID("00000000-0000-0000-0000-00000000e003")
    s3_expectations_store_backend_id = s3_expectations_store_backend.store_backend_id
    assert s3_expectations_store_backend_id != store_error_uuid

    # Make sure the bucket contents are as expected
    bucket_contents_after_creating_expectation_store = list_s3_bucket_contents(
        bucket=bucket, prefix=data_context_prefix
    )
    assert bucket_contents_after_creating_expectation_store == {
        f"{expectations_store_prefix}/{store_backend_id_filename}"
    }

    # Make sure the store_backend_id from the file is equal to reading from the property
    expectations_store_backend_id_from_s3_file = get_store_backend_id_from_s3(
        bucket=bucket,
        prefix=expectations_store_prefix,
        key=store_backend_id_filename,
    )
    assert expectations_store_backend_id_from_s3_file == s3_expectations_store_backend_id

    # Create a DataContext (note existing expectations store already set up)
    in_code_data_context_project_config = build_in_code_data_context_project_config(
        bucket="leakybucket",
        expectations_store_prefix=expectations_store_prefix,
        validation_results_store_prefix=validation_results_store_prefix,
        data_docs_store_prefix=data_docs_store_prefix,
    )
    in_code_data_context = get_context(project_config=in_code_data_context_project_config)
    bucket_contents_after_instantiating_get_context = list_s3_bucket_contents(
        bucket=bucket, prefix=data_context_prefix
    )
    assert bucket_contents_after_instantiating_get_context == {
        f"{expectations_store_prefix}/{store_backend_id_filename}",
        f"{validation_results_store_prefix}/{store_backend_id_filename}",
    }

    # Make sure ids are consistent
    in_code_data_context_expectations_store_store_backend_id = in_code_data_context.stores[
        "expectations_S3_store"
    ].store_backend_id
    in_code_data_context_data_context_id = in_code_data_context.data_context_id
    constructed_data_context_id = in_code_data_context._construct_data_context_id()
    assert (
        in_code_data_context_expectations_store_store_backend_id
        == in_code_data_context_data_context_id
        == expectations_store_backend_id_from_s3_file
        == s3_expectations_store_backend_id
        == constructed_data_context_id
    )


@pytest.mark.aws_deps
@mock_s3
def test_DataContext_construct_data_context_id_uses_id_stored_in_DataContextConfig_if_no_configured_expectations_store(  # noqa: E501 # FIXME CoP
    monkeypatch, aws_credentials
):
    """
    What does this test and why?

    A DataContext should have an id. This ID should come from either:
    1. configured expectations store store_backend_id
    2. great_expectations.yml
    3. new generated id from DataContextConfig
    This test verifies that DataContext._construct_data_context_id
    uses the data_context_id from DataContextConfig when there is no configured expectations store
    when instantiating the DataContext,
    and also that this data_context_id is used to configure the expectations_store.store_backend_id
    """
    bucket = "leakybucket"
    expectations_store_prefix = "expectations_store_prefix"
    validation_results_store_prefix = "validation_results_store_prefix"
    data_docs_store_prefix = "data_docs_store_prefix"
    manually_created_uuid = uuid.UUID("00000000-0000-0000-0000-000000000eee")

    # Create a bucket in Moto's mock AWS environment
    conn = boto3.resource("s3", region_name="us-east-1")
    conn.create_bucket(Bucket=bucket)

    # Create a DataContext (note NO existing expectations store already set up)
    in_code_data_context_project_config = build_in_code_data_context_project_config(
        bucket="leakybucket",
        expectations_store_prefix=expectations_store_prefix,
        validation_results_store_prefix=validation_results_store_prefix,
        data_docs_store_prefix=data_docs_store_prefix,
    )
    # Manually set the data_context_id in the project_config
    in_code_data_context_project_config.data_context_id = manually_created_uuid
    in_code_data_context = get_context(project_config=in_code_data_context_project_config)

    # Make sure the manually set data_context_id is propagated to all the appropriate places
    assert (
        manually_created_uuid
        == in_code_data_context.data_context_id
        == in_code_data_context.stores[
            in_code_data_context.expectations_store_name
        ].store_backend_id
    )


@pytest.mark.big
@mock_s3
def test_suppress_store_backend_id_is_true_for_inactive_stores():
    """
    What does this test and why?

    Trying to read / set the store_backend_id for inactive stores should not be attempted during DataContext initialization. This test ensures that the _suppress_store_backend_id parameter is set to True for inactive stores.

    """  # noqa: E501 # FIXME CoP

    bucket = "leakybucket"
    expectations_store_prefix = "expectations_store_prefix"
    validation_results_store_prefix = "validation_results_store_prefix"
    data_docs_store_prefix = "data_docs_store_prefix"

    # Create a bucket in Moto's mock AWS environment
    conn = boto3.resource("s3", region_name="us-east-1")
    conn.create_bucket(Bucket=bucket)

    # Create a DataContext
    # Add inactive stores
    inactive_bucket = "inactive_leakybucket"
    stores = {
        "expectations_S3_store": {
            "class_name": "ExpectationsStore",
            "store_backend": {
                "class_name": "TupleS3StoreBackend",
                "bucket": bucket,
                "prefix": expectations_store_prefix,
            },
        },
        "validation_results_S3_store": {
            "class_name": "ValidationResultsStore",
            "store_backend": {
                "class_name": "TupleS3StoreBackend",
                "bucket": bucket,
                "prefix": validation_results_store_prefix,
            },
        },
        "inactive_expectations_S3_store": {
            "class_name": "ExpectationsStore",
            "store_backend": {
                "class_name": "TupleS3StoreBackend",
                "bucket": inactive_bucket,
                "prefix": expectations_store_prefix,
            },
        },
        "inactive_validation_results_S3_store": {
            "class_name": "ValidationResultsStore",
            "store_backend": {
                "class_name": "TupleS3StoreBackend",
                "bucket": inactive_bucket,
                "prefix": validation_results_store_prefix,
            },
        },
    }
    in_code_data_context_project_config = build_in_code_data_context_project_config(
        bucket="leakybucket",
        expectations_store_prefix=expectations_store_prefix,
        validation_results_store_prefix=validation_results_store_prefix,
        data_docs_store_prefix=data_docs_store_prefix,
        stores=stores,
    )
    in_code_data_context = get_context(project_config=in_code_data_context_project_config)

    # Check here that suppress_store_backend_id == True for inactive stores
    # and False for active stores
    assert (
        in_code_data_context.stores.get(
            "inactive_expectations_S3_store"
        ).store_backend._suppress_store_backend_id
        is True
    )
    assert (
        in_code_data_context.stores.get(
            "inactive_validation_results_S3_store"
        ).store_backend._suppress_store_backend_id
        is True
    )
    assert (
        in_code_data_context.stores.get(
            "expectations_S3_store"
        ).store_backend._suppress_store_backend_id
        is False
    )
    assert (
        in_code_data_context.stores.get(
            "validation_results_S3_store"
        ).store_backend._suppress_store_backend_id
        is False
    )


@pytest.mark.aws_deps
@mock_s3
def test_inaccessible_active_bucket_warning_messages(caplog, aws_credentials):
    """
    What does this test do and why?

    Trying to create a data context with unreachable ACTIVE stores should show a warning message once per store
    e.g. Invalid store configuration: Please check the configuration of your TupleS3StoreBackend named expectations_S3_store
    Active stores are those named in:
    "expectations_store_name", "validation_results_store_name"
    """  # noqa: E501 # FIXME CoP

    bucket = "leakybucket"
    expectations_store_prefix = "expectations_store_prefix"
    validation_results_store_prefix = "validation_results_store_prefix"
    data_docs_store_prefix = "data_docs_store_prefix"

    # Create a bucket in Moto's mock AWS environment
    conn = boto3.resource("s3", region_name="us-east-1")
    conn.create_bucket(Bucket=bucket)

    # Create a DataContext
    # Add inactive stores
    inactive_bucket = "inactive_leakybucket"
    stores = {
        "expectations_S3_store": {
            "class_name": "ExpectationsStore",
            "store_backend": {
                "class_name": "TupleS3StoreBackend",
                "bucket": inactive_bucket,
                "prefix": expectations_store_prefix,
            },
        },
        "validation_results_S3_store": {
            "class_name": "ValidationResultsStore",
            "store_backend": {
                "class_name": "TupleS3StoreBackend",
                "bucket": inactive_bucket,
                "prefix": validation_results_store_prefix,
            },
        },
    }
    in_code_data_context_project_config = build_in_code_data_context_project_config(
        bucket="leakybucket",
        expectations_store_prefix=expectations_store_prefix,
        validation_results_store_prefix=validation_results_store_prefix,
        data_docs_store_prefix=data_docs_store_prefix,
        stores=stores,
    )
    _ = get_context(project_config=in_code_data_context_project_config)
    assert (
        caplog.messages.count(
            "Invalid store configuration: Please check the configuration of your TupleS3StoreBackend named expectations_S3_store. Exception was: \n Unable to set object in s3."  # noqa: E501 # FIXME CoP
        )
        == 1
    )
    assert (
        caplog.messages.count(
            "Invalid store configuration: Please check the configuration of your TupleS3StoreBackend named validation_results_S3_store. Exception was: \n Unable to set object in s3."  # noqa: E501 # FIXME CoP
        )
        == 1
    )


@pytest.mark.big
@mock_s3
def test_inaccessible_inactive_bucket_no_warning_messages(caplog):
    """
    What does this test do and why?

    Trying to create a data context with unreachable INACTIVE stores should show no warning messages
    Inactive stores are those NOT named in:
    "expectations_store_name", "validation_results_store_name"
    """

    bucket = "leakybucket"
    expectations_store_prefix = "expectations_store_prefix"
    validation_results_store_prefix = "validation_results_store_prefix"
    data_docs_store_prefix = "data_docs_store_prefix"

    # Create a bucket in Moto's mock AWS environment
    conn = boto3.resource("s3", region_name="us-east-1")
    conn.create_bucket(Bucket=bucket)

    # Create a DataContext
    # Add inactive stores
    inactive_bucket = "inactive_leakybucket"
    stores = {
        "expectations_S3_store": {
            "class_name": "ExpectationsStore",
            "store_backend": {
                "class_name": "TupleS3StoreBackend",
                "bucket": bucket,
                "prefix": expectations_store_prefix,
            },
        },
        "validation_results_S3_store": {
            "class_name": "ValidationResultsStore",
            "store_backend": {
                "class_name": "TupleS3StoreBackend",
                "bucket": bucket,
                "prefix": validation_results_store_prefix,
            },
        },
        "inactive_expectations_S3_store": {
            "class_name": "ExpectationsStore",
            "store_backend": {
                "class_name": "TupleS3StoreBackend",
                "bucket": inactive_bucket,
                "prefix": expectations_store_prefix,
            },
        },
        "inactive_validation_results_S3_store": {
            "class_name": "ValidationResultsStore",
            "store_backend": {
                "class_name": "TupleS3StoreBackend",
                "bucket": inactive_bucket,
                "prefix": validation_results_store_prefix,
            },
        },
    }
    in_code_data_context_project_config = build_in_code_data_context_project_config(
        bucket="leakybucket",
        expectations_store_prefix=expectations_store_prefix,
        validation_results_store_prefix=validation_results_store_prefix,
        data_docs_store_prefix=data_docs_store_prefix,
        stores=stores,
    )
    _ = get_context(project_config=in_code_data_context_project_config)
    assert (
        caplog.messages.count(
            "Invalid store configuration: Please check the configuration of your TupleS3StoreBackend named expectations_S3_store"  # noqa: E501 # FIXME CoP
        )
        == 0
    )
    assert (
        caplog.messages.count(
            "Invalid store configuration: Please check the configuration of your TupleS3StoreBackend named validation_results_S3_store"  # noqa: E501 # FIXME CoP
        )
        == 0
    )



"""Test v3 API data connector serialization."""

import pytest

from great_expectations.data_context.types.base import (
    DataConnectorConfig,
    dataConnectorConfigSchema,
)


@pytest.mark.unit
@pytest.mark.parametrize(
    "data_connector_config,expected_serialized_data_connector_config",
    [
        pytest.param(
            DataConnectorConfig(
                class_name="RuntimeDataConnector",
                batch_identifiers=["default_identifier_name"],
            ),
            {
                "class_name": "RuntimeDataConnector",
                "module_name": None,
                "batch_identifiers": ["default_identifier_name"],
            },
            id="no id",
        ),
        pytest.param(
            DataConnectorConfig(
                class_name="RuntimeDataConnector",
                id="dd8fe6df-254b-4e37-9c0e-2c8205d1e988",
                batch_identifiers=["default_identifier_name"],
            ),
            {
                "class_name": "RuntimeDataConnector",
                "module_name": None,
                "id": "dd8fe6df-254b-4e37-9c0e-2c8205d1e988",
                "batch_identifiers": ["default_identifier_name"],
            },
            id="with id",
        ),
    ],
)
def test_data_connector_config_is_serialized(
    data_connector_config: DataConnectorConfig,
    expected_serialized_data_connector_config: dict,
):
    """DataConnector Config should be serialized appropriately with/without optional params."""
    observed_dump = dataConnectorConfigSchema.dump(data_connector_config)
    assert observed_dump == expected_serialized_data_connector_config

    observed_load = dataConnectorConfigSchema.load(observed_dump)
    assert observed_load.to_json_dict() == data_connector_config.to_json_dict()
